"""Tests that perform actions over content unit."""

import os
from tempfile import NamedTemporaryFile

import pytest
import requests

from pulpcore.client.pulpcore.exceptions import BadRequestException
from pulpcore.client.pulp_rpm import ApiException
from pulpcore.tests.functional.utils import PulpTaskError
from pulp_rpm.tests.functional.constants import (
    BIG_COMPS_XML,
    BIG_CATEGORY,
    BIG_GROUPS,
    BIG_ENVIRONMENTS,
    BIG_LANGPACK,
    RPM_PACKAGEENVIRONMENT_CONTENT_NAME,
    RPM_PACKAGECATEGORY_CONTENT_NAME,
    RPM_PACKAGEGROUP_CONTENT_NAME,
    RPM_PACKAGELANGPACKS_CONTENT_NAME,
    RPM_UNSIGNED_FIXTURE_URL,
    RPM_PACKAGE_FILENAME,
    RPM_PACKAGE_FILENAME2,
    RPM_WITH_NON_ASCII_URL,
    SMALL_COMPS_XML,
    SMALL_CATEGORY,
    SMALL_ENVIRONMENTS,
    SMALL_GROUPS,
    SMALL_LANGPACK,
)

SMALL_CONTENT = SMALL_GROUPS + SMALL_CATEGORY + SMALL_LANGPACK + SMALL_ENVIRONMENTS
CENTOS8_CONTENT = BIG_GROUPS + BIG_CATEGORY + BIG_LANGPACK + BIG_ENVIRONMENTS


def test_single_request_unit_and_duplicate_unit(
    delete_orphans_pre, rpm_package_api, monitor_task, pulpcore_bindings
):
    """Test single request upload unit.

    1. Upload a unit
    2. Attempt to upload same unit
    """
    # Single unit upload
    file_to_use = os.path.join(RPM_UNSIGNED_FIXTURE_URL, RPM_PACKAGE_FILENAME)

    labels = {"key_1": "value_1"}
    with NamedTemporaryFile() as file_to_upload:
        file_to_upload.write(requests.get(file_to_use).content)
        upload_attrs = {"file": file_to_upload.name, "pulp_labels": labels}
        upload = rpm_package_api.create(**upload_attrs)

    content = monitor_task(upload.task).created_resources[0]
    package = rpm_package_api.read(content)
    assert package.location_href == RPM_PACKAGE_FILENAME
    assert package.pulp_labels == labels

    # Duplicate unit
    with NamedTemporaryFile() as file_to_upload:
        file_to_upload.write(requests.get(file_to_use).content)
        upload_attrs = {"file": file_to_upload.name}
        upload = rpm_package_api.create(**upload_attrs)

    try:
        monitor_task(upload.task)
    except PulpTaskError:
        pass
    task_report = pulpcore_bindings.TasksApi.read(upload.task)
    assert task_report.created_resources[0] == package.pulp_href


def test_upload_non_ascii(delete_orphans_pre, rpm_package_api, monitor_task):
    """Test whether one can upload an RPM with non-ascii metadata."""
    packages_count = rpm_package_api.list().count
    with NamedTemporaryFile() as file_to_upload:
        file_to_upload.write(requests.get(RPM_WITH_NON_ASCII_URL).content)
        upload_attrs = {"file": file_to_upload.name}
        upload = rpm_package_api.create(**upload_attrs)

    monitor_task(upload.task)
    new_packages_count = rpm_package_api.list().count
    assert (packages_count + 1) == new_packages_count


@pytest.fixture
def upload_comps_into(rpm_comps_api, monitor_task):
    def _upload_comps_into(file_path, expected_totals, repo_href=None, replace=False):
        data = {}
        if repo_href:
            data["repository"] = repo_href
            data["replace"] = replace
            expected_totals += 1
        response = rpm_comps_api.rpm_comps_upload(file=file_path, **data)
        task = monitor_task(response.task)
        rsrcs = task.created_resources
        assert expected_totals == len(rsrcs)
        return rsrcs

    return _upload_comps_into


def test_upload_comps_xml(delete_orphans_pre, upload_comps_into):
    """Upload a comps.xml and make sure it created comps-Content."""
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, SMALL_CONTENT)
    eval_resources(resources)


def test_upload_same_comps_xml(delete_orphans_pre, upload_comps_into):
    """Upload a comps.xml twice and make sure it doesn't create new objects the second time."""
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        first = upload_comps_into(comps_file.name, SMALL_CONTENT)
        second = upload_comps_into(comps_file.name, SMALL_CONTENT)

    # we return all resources in the comps.xml, even if they already existed
    assert sorted(first) == sorted(second)


def test_upload_diff_comps_xml(delete_orphans_pre, upload_comps_into):
    """Upload two different comps-files and make sure the second shows the new comps-Content."""
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        upload_comps_into(comps_file.name, SMALL_CONTENT)
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(BIG_COMPS_XML)
        comps_file.flush()
        second = upload_comps_into(comps_file.name, CENTOS8_CONTENT)
    eval_resources(second, is_small=False)


def test_upload_comps_xml_into_repo(
    delete_orphans_pre, rpm_repository_factory, rpm_repository_version_api, upload_comps_into
):
    """Upload comps into a repo and see new version created containing the comps-Content."""
    repo = rpm_repository_factory()

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, SMALL_CONTENT, repo.pulp_href)

    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 3
    eval_counts(vers_resp.content_summary.added)


def test_upload_comps_xml_into_repo_add(
    delete_orphans_pre, rpm_repository_factory, rpm_repository_version_api, upload_comps_into
):
    """Upload two comps-files into a repo and see the result being additive."""
    repo = rpm_repository_factory()

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, SMALL_CONTENT, repo.pulp_href)

    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 3

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(BIG_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, CENTOS8_CONTENT, repo.pulp_href)
    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert vers_resp.number == 2

    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 4
    eval_counts(vers_resp.content_summary.added, is_small=False)
    eval_sum_counts(vers_resp.content_summary.present)


def test_upload_comps_xml_into_repo_replace(
    delete_orphans_pre, rpm_repository_factory, rpm_repository_version_api, upload_comps_into
):
    """Upload two comps, see the comps-content from the second replace the existing."""
    repo = rpm_repository_factory()

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        upload_comps_into(comps_file.name, SMALL_CONTENT, repo.pulp_href)
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(BIG_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, CENTOS8_CONTENT, repo.pulp_href, True)
    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert vers_resp.number == 2

    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 3
    eval_counts(vers_resp.content_summary.added, is_small=False)


def test_synchronous_package_upload(delete_orphans_pre, rpm_package_api, gen_user):
    """Test synchronously uploading an RPM.

    1. Upload a unit
    2. Attempt to upload same unit with different labels
    3. Assert that labels don't change.
    """
    # Single unit upload
    file_to_use = os.path.join(RPM_UNSIGNED_FIXTURE_URL, RPM_PACKAGE_FILENAME)

    with gen_user(model_roles=["rpm.rpm_package_uploader"]):
        labels = {"key_1": "value_1"}
        with NamedTemporaryFile() as file_to_upload:
            file_to_upload.write(requests.get(file_to_use).content)
            upload_attrs = {"file": file_to_upload.name, "pulp_labels": labels}
            package = rpm_package_api.upload(**upload_attrs)

        assert package.location_href == RPM_PACKAGE_FILENAME
        assert package.pulp_labels == labels

        # Duplicate unit
        with NamedTemporaryFile() as file_to_upload:
            new_labels = {"key_2": "value_2"}
            file_to_upload.write(requests.get(file_to_use).content)
            upload_attrs = {"file": file_to_upload.name, "pulp_labels": new_labels}
            duplicate_package = rpm_package_api.upload(**upload_attrs)

        assert duplicate_package.pulp_href == package.pulp_href
        assert duplicate_package.pulp_labels == package.pulp_labels
        assert duplicate_package.pulp_labels != new_labels

    with gen_user(model_roles=[]), pytest.raises(ApiException) as ctx:
        labels = {"key_1": "value_1"}
        with NamedTemporaryFile() as file_to_upload:
            file_to_upload.write(requests.get(file_to_use).content)
            upload_attrs = {"file": file_to_upload.name, "pulp_labels": labels}
            rpm_package_api.upload(**upload_attrs)
    assert ctx.value.status == 403


def test_synchronous_package_upload_from_artifact(rpm_package_api, gen_user, pulpcore_bindings):
    """Test synchronously uploading an RPM.

    1. Create an Artifact from an RPM, if it doesn't exist.
    2. Use synchronous RPM upload API with an Artifact.
    3. Assert that the RPM package created has a matching artifact.
    """
    file_to_use = os.path.join(RPM_UNSIGNED_FIXTURE_URL, RPM_PACKAGE_FILENAME2)
    with NamedTemporaryFile() as file_to_upload:
        file_to_upload.write(requests.get(file_to_use).content)
        try:
            artifact = pulpcore_bindings.ArtifactsApi.create(file_to_upload.name)
        except BadRequestException as exc:
            sha256sum = exc.body.split("'")[1]
            artifact = pulpcore_bindings.ArtifactsApi.list(sha256=sha256sum).results[0]

    with gen_user(model_roles=["rpm.rpm_package_uploader"]):
        # Using an existing artifact
        upload_attrs = {"artifact": artifact.pulp_href}
        package_from_artifact = rpm_package_api.upload(**upload_attrs)
    assert package_from_artifact.artifact == artifact.pulp_href


def test_synchronous_package_upload_from_chunks(
    delete_orphans_pre, rpm_package_api, gen_user, pulpcore_bindings, tmp_path
):
    """Test synchronously uploading an RPM using chunked upload.

    1. Upload an RPM file in chunks.
    2. Use synchronous RPM upload API with the upload object.
    3. Assert that the RPM package is created successfully.
    """
    import hashlib
    import uuid

    file_to_use = os.path.join(RPM_UNSIGNED_FIXTURE_URL, RPM_PACKAGE_FILENAME)

    # Download the file and prepare chunks
    with NamedTemporaryFile(delete=False) as file_to_upload:
        file_to_upload.write(requests.get(file_to_use).content)
        file_to_upload.flush()
        file_path = file_to_upload.name

    # Create chunks (similar to pulpcore_chunked_file_factory)
    chunk_size = 512
    chunks = []
    hasher = hashlib.new("sha256")

    with open(file_path, "rb") as f:
        data = f.read()

    file_size = len(data)
    start = 0

    while start < len(data):
        content = data[start : start + chunk_size]
        chunk_file = tmp_path / str(uuid.uuid4())
        hasher.update(content)
        chunk_file.write_bytes(content)
        content_sha = hashlib.sha256(content).hexdigest()
        end = start + len(content) - 1
        chunks.append((str(chunk_file), f"bytes {start}-{end}/{file_size}", content_sha))
        start += len(content)

    # Create an Upload object (requires core permissions, done outside gen_user)
    upload = pulpcore_bindings.UploadsApi.create({"size": file_size})

    # Upload all chunks
    for chunk_file, content_range, chunk_sha in chunks:
        pulpcore_bindings.UploadsApi.update(
            upload_href=upload.pulp_href,
            file=chunk_file,
            content_range=content_range,
            sha256=chunk_sha,
        )

    # Use synchronous upload API with the upload object (requires rpm uploader role)
    with gen_user(model_roles=["rpm.rpm_package_uploader"]):
        upload_attrs = {"upload": upload.pulp_href}
        package = rpm_package_api.upload(**upload_attrs)

        # Verify package was created successfully
        assert package.location_href == RPM_PACKAGE_FILENAME

    # Clean up
    os.unlink(file_path)


def eval_resources(resources, is_small=True):
    """Eval created_resources counts."""
    groups = [g for g in resources if "packagegroups" in g]
    assert len(groups) == (SMALL_GROUPS if is_small else BIG_GROUPS)

    categories = [g for g in resources if "packagecategories" in g]
    assert len(categories) == (SMALL_CATEGORY if is_small else BIG_CATEGORY)

    langpacks = [g for g in resources if "packagelangpacks" in g]
    assert len(langpacks) == (SMALL_LANGPACK if is_small else BIG_LANGPACK)

    envs = [g for g in resources if "environment" in g]
    assert len(envs) == (SMALL_ENVIRONMENTS if is_small else BIG_ENVIRONMENTS)


def eval_counts(summary, is_small=True):
    """Eval counts in a given summary."""
    for c in summary:
        if "rpm.packagegroup" in c:
            count = SMALL_GROUPS if is_small else BIG_GROUPS
            assert summary["rpm.packagegroup"]["count"] == count
        elif "rpm.packageenvironment" in c:
            count = SMALL_ENVIRONMENTS if is_small else BIG_ENVIRONMENTS
            assert summary["rpm.packageenvironment"]["count"] == count
        elif RPM_PACKAGECATEGORY_CONTENT_NAME in c:
            count = SMALL_CATEGORY if is_small else BIG_CATEGORY
            assert summary[RPM_PACKAGECATEGORY_CONTENT_NAME]["count"] == count
        elif "rpm.packagelangpacks" in c:
            count = SMALL_LANGPACK if is_small else BIG_LANGPACK
            assert summary["rpm.packagelangpacks"]["count"] == count


def eval_sum_counts(summary):
    """In a given summary, counts should be BIG+SMALL."""
    for c in summary:
        if RPM_PACKAGEGROUP_CONTENT_NAME in c:
            count = BIG_GROUPS + SMALL_GROUPS
            assert summary[RPM_PACKAGEGROUP_CONTENT_NAME]["count"] == count
        elif RPM_PACKAGEENVIRONMENT_CONTENT_NAME in c:
            count = BIG_ENVIRONMENTS + SMALL_ENVIRONMENTS
            assert summary[RPM_PACKAGEENVIRONMENT_CONTENT_NAME]["count"] == count
        elif RPM_PACKAGECATEGORY_CONTENT_NAME in c:
            count = BIG_CATEGORY + SMALL_CATEGORY
            assert summary[RPM_PACKAGECATEGORY_CONTENT_NAME]["count"] == count
        elif RPM_PACKAGELANGPACKS_CONTENT_NAME in c:
            count = BIG_LANGPACK + SMALL_LANGPACK
            assert summary[RPM_PACKAGELANGPACKS_CONTENT_NAME]["count"] == count



"""
Unit tests for Package age calculation functionality.
"""

from django.test import TestCase

from pulp_rpm.app.models import Package
from pulp_rpm.app.shared_utils import annotate_with_age


class TestPackageAge(TestCase):
    """Test Package age calculation functionality."""

    def setUp(self):
        """Set up test packages with various version combinations."""
        # Package group 1: same name and arch, different versions
        self.pkg1_v1 = Package.objects.create(
            name="testpkg",
            epoch="0",
            version="1.0.0",
            release="1.el8",
            arch="x86_64",
            pkgId="checksum1",
            checksum_type="sha256",
        )

        self.pkg1_v2 = Package.objects.create(
            name="testpkg",
            epoch="0",
            version="2.0.0",
            release="1.el8",
            arch="x86_64",
            pkgId="checksum2",
            checksum_type="sha256",
        )

        self.pkg1_v3 = Package.objects.create(
            name="testpkg",
            epoch="0",
            version="1.5.0",
            release="2.el8",
            arch="x86_64",
            pkgId="checksum3",
            checksum_type="sha256",
        )

        # Package group 2: same name but different arch
        self.pkg2_i686 = Package.objects.create(
            name="testpkg",
            epoch="0",
            version="1.0.0",
            release="1.el8",
            arch="i686",
            pkgId="checksum4",
            checksum_type="sha256",
        )

        self.pkg2_i686_v2 = Package.objects.create(
            name="testpkg",
            epoch="0",
            version="3.0.0",
            release="1.el8",
            arch="i686",
            pkgId="checksum5",
            checksum_type="sha256",
        )

        # Package group 3: different name
        self.pkg3_other = Package.objects.create(
            name="otherpkg",
            epoch="0",
            version="1.0.0",
            release="1.el8",
            arch="x86_64",
            pkgId="checksum6",
            checksum_type="sha256",
        )

        # Package group 4: epoch differences
        self.pkg4_epoch0 = Package.objects.create(
            name="epochpkg",
            epoch="0",
            version="2.0.0",
            release="1.el8",
            arch="x86_64",
            pkgId="checksum7",
            checksum_type="sha256",
        )

        self.pkg4_epoch1 = Package.objects.create(
            name="epochpkg",
            epoch="1",
            version="1.0.0",
            release="1.el8",
            arch="x86_64",
            pkgId="checksum8",
            checksum_type="sha256",
        )

    def test_age_calculation_basic_versions(self):
        """Test that age is calculated correctly for basic version differences."""
        packages = annotate_with_age(
            Package.objects.filter(name="testpkg", arch="x86_64")
        ).order_by("age")

        # Expected order: 2.0.0 (age=1), 1.5.0-2 (age=2), 1.0.0 (age=3)
        self.assertEqual(packages.count(), 3)

        # Newest version should have age=1
        newest = packages[0]
        self.assertEqual(newest.version, "2.0.0")
        self.assertEqual(newest.age, 1)

        # Middle version should have age=2
        middle = packages[1]
        self.assertEqual(middle.version, "1.5.0")
        self.assertEqual(middle.release, "2.el8")
        self.assertEqual(middle.age, 2)

        # Oldest version should have age=3
        oldest = packages[2]
        self.assertEqual(oldest.version, "1.0.0")
        self.assertEqual(oldest.age, 3)

    def test_age_calculation_different_architectures(self):
        """Test that packages with different architectures are aged separately."""
        # x86_64 packages
        x86_packages = annotate_with_age(
            Package.objects.filter(name="testpkg", arch="x86_64")
        ).order_by("age")
        self.assertEqual(x86_packages.count(), 3)
        self.assertEqual(x86_packages[0].age, 1)  # 2.0.0
        self.assertEqual(x86_packages[1].age, 2)  # 1.5.0-2
        self.assertEqual(x86_packages[2].age, 3)  # 1.0.0

        # i686 packages
        i686_packages = annotate_with_age(
            Package.objects.filter(name="testpkg", arch="i686")
        ).order_by("age")
        self.assertEqual(i686_packages.count(), 2)

        # Even though i686 3.0.0 is newer than any x86_64 version, it should still have age=1
        # within its own architecture group
        self.assertEqual(i686_packages[0].version, "3.0.0")
        self.assertEqual(i686_packages[0].age, 1)
        self.assertEqual(i686_packages[1].version, "1.0.0")
        self.assertEqual(i686_packages[1].age, 2)

    def test_age_calculation_different_names(self):
        """Test that packages with different names are aged separately."""
        # Different package name should have its own age calculation
        other_packages = annotate_with_age(Package.objects.filter(name="otherpkg"))
        self.assertEqual(other_packages.count(), 1)
        self.assertEqual(other_packages[0].age, 1)

    def test_age_calculation_with_epochs(self):
        """Test that epoch is considered in age calculation."""
        epoch_packages = annotate_with_age(Package.objects.filter(name="epochpkg")).order_by("age")
        self.assertEqual(epoch_packages.count(), 2)

        # Epoch 1 should be newer than epoch 0, regardless of version numbers
        newest = epoch_packages[0]
        self.assertEqual(newest.epoch, "1")
        self.assertEqual(newest.version, "1.0.0")
        self.assertEqual(newest.age, 1)

        oldest = epoch_packages[1]
        self.assertEqual(oldest.epoch, "0")
        self.assertEqual(oldest.version, "2.0.0")
        self.assertEqual(oldest.age, 2)

    def test_age_all_packages(self):
        """Test age calculation when querying all packages."""
        all_packages = annotate_with_age(Package.objects.all())

        # Verify each package has an age
        for pkg in all_packages:
            self.assertIsNotNone(pkg.age)
            self.assertGreater(pkg.age, 0)

        # Check that packages are grouped correctly by name+arch
        testpkg_x86_ages = [
            p.age for p in all_packages if p.name == "testpkg" and p.arch == "x86_64"
        ]
        testpkg_i686_ages = [
            p.age for p in all_packages if p.name == "testpkg" and p.arch == "i686"
        ]
        otherpkg_ages = [p.age for p in all_packages if p.name == "otherpkg"]
        epochpkg_ages = [p.age for p in all_packages if p.name == "epochpkg"]

        self.assertEqual(sorted(testpkg_x86_ages), [1, 2, 3])
        self.assertEqual(sorted(testpkg_i686_ages), [1, 2])
        self.assertEqual(sorted(otherpkg_ages), [1])
        self.assertEqual(sorted(epochpkg_ages), [1, 2])

    def test_age_with_release_differences(self):
        """Test age calculation when versions are same but releases differ."""
        # Create packages with same version but different releases
        rel_pkg1 = Package.objects.create(  # noqa: F841
            name="relpkg",
            epoch="0",
            version="1.0.0",
            release="1.el8",
            arch="x86_64",
            pkgId="rel1",
            checksum_type="sha256",
        )

        rel_pkg2 = Package.objects.create(  # noqa: F841
            name="relpkg",
            epoch="0",
            version="1.0.0",
            release="2.el8",
            arch="x86_64",
            pkgId="rel2",
            checksum_type="sha256",
        )

        rel_pkg3 = Package.objects.create(  # noqa: F841
            name="relpkg",
            epoch="0",
            version="1.0.0",
            release="10.el8",
            arch="x86_64",
            pkgId="rel3",
            checksum_type="sha256",
        )

        packages = annotate_with_age(Package.objects.filter(name="relpkg")).order_by("age")

        # Expected order: 10.el8 > 2.el8 > 1.el8 (numeric comparison of release)
        self.assertEqual(packages[0].release, "10.el8")  # age=1
        self.assertEqual(packages[0].age, 1)

        self.assertEqual(packages[1].release, "2.el8")  # age=2
        self.assertEqual(packages[1].age, 2)

        self.assertEqual(packages[2].release, "1.el8")  # age=3
        self.assertEqual(packages[2].age, 3)

    def test_age_with_tilde_and_caret_versions(self):
        """Test age calculation with tilde and caret version characters."""
        # Create packages with tilde and caret versions
        Package.objects.create(
            name="tildepkg",
            epoch="0",
            version="1.0~rc1",
            release="1.el8",
            arch="x86_64",
            pkgId="tilde1",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="tildepkg",
            epoch="0",
            version="1.0",
            release="1.el8",
            arch="x86_64",
            pkgId="tilde2",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="tildepkg",
            epoch="0",
            version="1.0~rc2",
            release="1.el8",
            arch="x86_64",
            pkgId="tilde3",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="tildepkg",
            epoch="0",
            version="1.0^git123",
            release="1.el8",
            arch="x86_64",
            pkgId="tilde4",
            checksum_type="sha256",
        )

        packages = annotate_with_age(Package.objects.filter(name="tildepkg")).order_by("age")

        # Expected order: 1.0^git123 (age=1), 1.0 (age=2), 1.0~rc2 (age=3), 1.0~rc1 (age=4)
        # Caret sorts higher than regular, regular sorts higher than tilde
        self.assertEqual(packages[0].version, "1.0^git123")
        self.assertEqual(packages[0].age, 1)

        self.assertEqual(packages[1].version, "1.0")
        self.assertEqual(packages[1].age, 2)

        self.assertEqual(packages[2].version, "1.0~rc2")
        self.assertEqual(packages[2].age, 3)

        self.assertEqual(packages[3].version, "1.0~rc1")
        self.assertEqual(packages[3].age, 4)

    def test_age_with_numeric_handling_versions(self):
        """Test age calculation with numeric version handling edge cases."""
        # Create packages with leading zeros and numeric comparisons
        Package.objects.create(
            name="numpkg",
            epoch="0",
            version="10.0001",
            release="1.el8",
            arch="x86_64",
            pkgId="num1",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="numpkg",
            epoch="0",
            version="10.1",
            release="1.el8",
            arch="x86_64",
            pkgId="num2",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="numpkg",
            epoch="0",
            version="10.10001",
            release="1.el8",
            arch="x86_64",
            pkgId="num3",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="numpkg",
            epoch="0",
            version="20240521",
            release="1.el8",
            arch="x86_64",
            pkgId="num4",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="numpkg",
            epoch="0",
            version="202405210",
            release="1.el8",
            arch="x86_64",
            pkgId="num5",
            checksum_type="sha256",
        )

        packages = annotate_with_age(Package.objects.filter(name="numpkg")).order_by("age")

        # Expected order: 202405210 > 20240521 > 10.10001 > 10.1 == 10.0001
        # Leading zeros are ignored in numeric segments
        self.assertEqual(packages[0].version, "202405210")
        self.assertEqual(packages[0].age, 1)

        self.assertEqual(packages[1].version, "20240521")
        self.assertEqual(packages[1].age, 2)

        self.assertEqual(packages[2].version, "10.10001")
        self.assertEqual(packages[2].age, 3)

        # 10.1 and 10.0001 should be considered equal (leading zeros ignored)
        # but one will have age=4 and the other age=5 based on creation order
        remaining_versions = [packages[3].version, packages[4].version]
        self.assertIn("10.1", remaining_versions)
        self.assertIn("10.0001", remaining_versions)

    def test_age_with_non_intuitive_comparison_versions(self):
        """Test age calculation with non-intuitive version comparison behavior."""
        # Create packages that test the 'e' vs numeric behavior
        Package.objects.create(
            name="intuitivepkg",
            epoch="0",
            version="1e.fc33",
            release="1.el8",
            arch="x86_64",
            pkgId="intuit1",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="intuitivepkg",
            epoch="0",
            version="1.fc33",
            release="1.el8",
            arch="x86_64",
            pkgId="intuit2",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="intuitivepkg",
            epoch="0",
            version="1g.fc33",
            release="1.el8",
            arch="x86_64",
            pkgId="intuit3",
            checksum_type="sha256",
        )

        packages = annotate_with_age(Package.objects.filter(name="intuitivepkg")).order_by("age")

        # Expected order: 1g.fc33 > 1.fc33 > 1e.fc33
        # 'e' comes before numeric in comparison, 'g' comes after numeric
        self.assertEqual(packages[0].version, "1g.fc33")
        self.assertEqual(packages[0].age, 1)

        self.assertEqual(packages[1].version, "1.fc33")
        self.assertEqual(packages[1].age, 2)

        self.assertEqual(packages[2].version, "1e.fc33")
        self.assertEqual(packages[2].age, 3)

    def test_age_with_non_alphanumeric_equivalence_versions(self):
        """Test age calculation with non-alphanumeric character equivalence."""
        # Create packages with various non-alphanumeric separators
        Package.objects.create(
            name="alphapkg",
            epoch="0",
            version="4.0",
            release="1.el8",
            arch="x86_64",
            pkgId="alpha1",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="alphapkg",
            epoch="0",
            version="4_0",
            release="1.el8",
            arch="x86_64",
            pkgId="alpha2",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="alphapkg",
            epoch="0",
            version="4+0",
            release="1.el8",
            arch="x86_64",
            pkgId="alpha3",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="alphapkg",
            epoch="0",
            version="4.999",
            release="1.el8",
            arch="x86_64",
            pkgId="alpha4",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="alphapkg",
            epoch="0",
            version="4.999.9",
            release="1.el8",
            arch="x86_64",
            pkgId="alpha5",
            checksum_type="sha256",
        )

        packages = annotate_with_age(Package.objects.filter(name="alphapkg")).order_by("age")

        # Expected behavior: 4.999.9 > 4.999 > 4.0 == 4_0 == 4+0
        # Non-alphanumeric characters are treated as equivalent separators
        self.assertEqual(packages[0].version, "4.999.9")
        self.assertEqual(packages[0].age, 1)

        self.assertEqual(packages[1].version, "4.999")
        self.assertEqual(packages[1].age, 2)

        # The remaining three should be considered equivalent versions
        # but will have different ages based on creation order
        remaining_versions = [packages[2].version, packages[3].version, packages[4].version]
        self.assertIn("4.0", remaining_versions)
        self.assertIn("4_0", remaining_versions)
        self.assertIn("4+0", remaining_versions)

    def test_age_with_non_ascii_character_versions(self):
        """Test age calculation with non-ASCII character handling."""
        # Create packages with non-ASCII characters
        Package.objects.create(
            name="asciipkg",
            epoch="0",
            version="1.1.1",
            release="1.el8",
            arch="x86_64",
            pkgId="ascii1",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="asciipkg",
            epoch="0",
            version="1.1.Á.1",
            release="1.el8",
            arch="x86_64",
            pkgId="ascii2",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="asciipkg",
            epoch="0",
            version="1.11",
            release="1.el8",
            arch="x86_64",
            pkgId="ascii3",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="asciipkg",
            epoch="0",
            version="1.1Á1",
            release="1.el8",
            arch="x86_64",
            pkgId="ascii4",
            checksum_type="sha256",
        )

        packages = annotate_with_age(Package.objects.filter(name="asciipkg")).order_by("age")

        # Expected behavior: 1.11 > 1.1.1 == 1.1.Á.1 > 1.1Á1
        # Non-ASCII chars are ignored unless they break up alphanumeric sequences
        self.assertEqual(packages[0].version, "1.11")
        self.assertEqual(packages[0].age, 1)

        # 1.1.1 and 1.1.Á.1 should be equivalent
        equivalent_versions = [packages[1].version, packages[2].version]
        self.assertIn("1.1.1", equivalent_versions)
        self.assertIn("1.1.Á.1", equivalent_versions)

        # 1.1Á1 should be last (non-ASCII breaks up the sequence)
        self.assertEqual(packages[3].version, "1.1Á1")
        self.assertEqual(packages[3].age, 4)

    def test_age_with_mixed_scenarios(self):
        """Test age calculation with mixed version scenarios."""
        Package.objects.create(
            name="mixedpkg",
            epoch="0",
            version="2.0",
            release="1.el8",
            arch="x86_64",
            pkgId="mixed1",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="mixedpkg",
            epoch="0",
            version="2.0.rc1",
            release="1.el8",
            arch="x86_64",
            pkgId="mixed2",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="mixedpkg",
            epoch="0",
            version="2.0.0",
            release="1.el8",
            arch="x86_64",
            pkgId="mixed3",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="mixedpkg",
            epoch="1",
            version="1.0",
            release="1.el8",
            arch="x86_64",
            pkgId="mixed4",
            checksum_type="sha256",
        )

        packages = annotate_with_age(Package.objects.filter(name="mixedpkg")).order_by("age")

        # Expected order: 1:1.0 (epoch wins) > 2.0.0 > 2.0 > 2.0.rc1
        self.assertEqual(packages[0].epoch, "1")
        self.assertEqual(packages[0].version, "1.0")
        self.assertEqual(packages[0].age, 1)

        # Rest should be ordered by version comparison
        remaining = packages[1:]
        versions = [p.version for p in remaining]
        self.assertIn("2.0.0", versions)
        self.assertIn("2.0", versions)
        self.assertIn("2.0.rc1", versions)

    def test_age_with_filtered_queryset(self):
        """Test age calculation when applied to a filtered queryset."""
        # Create packages in database
        pkg1 = Package.objects.create(  # noqa: F841
            name="filterpkg",
            epoch="0",
            version="1.0",
            release="1.el8",
            arch="x86_64",
            pkgId="filter1",
            checksum_type="sha256",
        )

        pkg2 = Package.objects.create(  # noqa: F841
            name="filterpkg",
            epoch="0",
            version="2.0",
            release="1.el8",
            arch="x86_64",
            pkgId="filter2",
            checksum_type="sha256",
        )

        Package.objects.create(
            name="filterpkg",
            epoch="0",
            version="3.0",
            release="1.el8",
            arch="x86_64",
            pkgId="filter3",
            checksum_type="sha256",
        )

        # Test with filtered queryset (only older versions)
        filtered_packages = annotate_with_age(
            Package.objects.filter(name="filterpkg", version__in=["1.0", "2.0"])
        ).order_by("age")

        self.assertEqual(filtered_packages.count(), 2)
        # In the filtered set, 2.0 should be newest (age=1), 1.0 should be oldest (age=2)
        self.assertEqual(filtered_packages[0].version, "2.0")
        self.assertEqual(filtered_packages[0].age, 1)
        self.assertEqual(filtered_packages[1].version, "1.0")
        self.assertEqual(filtered_packages[1].age, 2)

    def test_age_with_empty_queryset(self):
        """Test age calculation with empty queryset."""
        empty_packages = annotate_with_age(Package.objects.filter(name="nonexistent"))
        self.assertEqual(len(empty_packages.all()), 0)

    def test_age_with_mixed_filtering_scenarios(self):
        """Test age calculation with various filtering scenarios that might cause bugs."""
        # Create packages across multiple groups
        for i in range(5):
            Package.objects.create(
                name="grouppkg",
                epoch="0",
                version=f"{i}.0",
                release="1.el8",
                arch="x86_64",
                pkgId=f"group{i}",
                checksum_type="sha256",
            )

        # Create another group
        for i in range(3):
            Package.objects.create(
                name="otherpkg",
                epoch="0",
                version=f"{i}.0",
                release="1.el8",
                arch="x86_64",
                pkgId=f"other{i}",
                checksum_type="sha256",
            )

        # Test 1: Filter to get partial group from first package set
        partial_group = annotate_with_age(
            Package.objects.filter(name="grouppkg", version__in=["2.0", "3.0", "4.0"])
        ).order_by("age")

        self.assertEqual(partial_group.count(), 3)
        # In this filtered set: 4.0 (age=1), 3.0 (age=2), 2.0 (age=3)
        self.assertEqual(partial_group[0].version, "4.0")
        self.assertEqual(partial_group[0].age, 1)
        self.assertEqual(partial_group[1].version, "3.0")
        self.assertEqual(partial_group[1].age, 2)
        self.assertEqual(partial_group[2].version, "2.0")
        self.assertEqual(partial_group[2].age, 3)

        # Test 2: Filter across multiple package groups
        cross_group = annotate_with_age(
            Package.objects.filter(name__in=["grouppkg", "otherpkg"], version="2.0")
        )

        # Should have one package from each group, both with age=1 in their respective groups
        for pkg in cross_group:
            self.assertEqual(pkg.version, "2.0")
            self.assertEqual(pkg.age, 1)  # Each is newest in its filtered group

    def test_age_consistency_with_retention_scenario(self):
        """Test age calculation in a scenario similar to how retention policies work"""
        packages_data = [
            ("retentiontestpkg", "1.0", "1.el8"),
            ("retentiontestpkg", "1.1", "1.el8"),
            ("retentiontestpkg", "1.2", "1.el8"),
            ("retentiontestpkg", "2.0", "1.el8"),
            ("retentiontestpkg", "2.1", "1.el8"),
        ]

        created_packages = []
        for i, (name, version, release) in enumerate(packages_data):
            pkg = Package.objects.create(
                name=name,
                epoch="0",
                version=version,
                release=release,
                arch="x86_64",
                pkgId=f"retention{i}",
                checksum_type="sha256",
            )
            created_packages.append(pkg)

        # Test: Get packages and apply age - this should match what retention logic does
        all_packages = annotate_with_age(Package.objects.filter(name="retentiontestpkg"))

        # Verify all packages have correct age values
        for pkg in all_packages:
            self.assertIsNotNone(pkg.age)
            self.assertGreater(pkg.age, 0)

        # Test different retention scenarios
        # Scenario 1: Keep only newest 2 versions (should remove 3 packages)
        oldest_packages = all_packages.filter(age__gt=2)
        self.assertEqual(oldest_packages.count(), 3)

        # Scenario 2: Keep only newest 3 versions (should remove 2 packages)
        oldest_packages = all_packages.filter(age__gt=3)
        self.assertEqual(oldest_packages.count(), 2)



# Sourced from https://github.com/nexB/univers
#
# Copyright (c) SAS Institute Inc.
# Copyright (c) Facebook, Inc. and its affiliates.
#
# SPDX-License-Identifier: MIT AND Apache-2.0
# Version comparison utility extracted from python-rpm-vercmp and further
# stripped down and significantly modified from the original at python-rpm-vercmp
# Also includes updates from Facebook antlir merged in.
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

# flake8: noqa

from typing import NamedTuple
from typing import Union


class RpmVersion(NamedTuple):
    """
    Represent an RPM version. It is ordered.
    """

    epoch: str
    version: str
    release: str

    def __str__(self, *args, **kwargs):
        return self.to_string()

    def to_string(self):
        if self.release:
            vr = f"{self.version}-{self.release}"
        else:
            vr = self.version

        if self.epoch:
            vr = f"{self.epoch}:{vr}"
        return vr

    @classmethod
    def from_string(cls, s):
        s.strip()
        e, v, r = from_evr(s)
        return cls(e, v, r)

    def __lt__(self, other):
        return _compare_rpm_versions(self, other) < 0

    def __gt__(self, other):
        return _compare_rpm_versions(self, other) > 0

    def __eq__(self, other):
        return _compare_rpm_versions(self, other) == 0

    def __le__(self, other):
        return _compare_rpm_versions(self, other) <= 0

    def __ge__(self, other):
        return _compare_rpm_versions(self, other) >= 0


def from_evr(s):
    """
    Return an (E, V, R) tuple given a string by splitting
    [e:]version-release into the three possible subcomponents.
    Default epoch, version and release to empty string if not specified.

    >>> assert from_evr("1:11.13.2.0-1") == ("1", "11.13.2.0", "1")
    >>> assert from_evr("11.13.2.0-1") == ("", "11.13.2.0", "1")
    """
    if ":" in s:
        e, _, vr = s.partition(":")
    else:
        e = ""
        vr = s

    if "-" in vr:
        v, _, r = vr.partition("-")
    else:
        v = vr
        r = ""
    return e, v, r


def _compare_rpm_versions(a: Union[RpmVersion, str], b: Union[RpmVersion, str]) -> int:
    """
    Compare two RPM versions ``a`` and ``b`` and return:
    -  1 if the version of a is newer than b
    -  0 if the versions match
    -  -1 if the version of a is older than b

    These are the legacy "cmp()" function semantics.

    This implementation is adapted from both this blog post:
    https://blog.jasonantman.com/2014/07/how-yum-and-rpm-compare-versions/
    and this Apache 2 licensed implementation:
    https://github.com/sassoftware/python-rpm-vercmp/blob/master/rpm_vercmp/vercmp.py

    For example::
    >>> assert compare_rpm_versions("1.0", "1.1") == -1
    >>> assert compare_rpm_versions("1.1", "1.0") == 1
    >>> assert compare_rpm_versions("11.13.2-1", "11.13.2.0-1") == -1
    >>> assert compare_rpm_versions("11.13.2.0-1", "11.13.2-1") == 1
    """
    if isinstance(a, str):
        a = RpmVersion.from_string(a)
    if isinstance(b, str):
        b = RpmVersion.from_string(b)
    if not isinstance(a, RpmVersion) and not isinstance(b, RpmVersion):
        raise TypeError(f"{a!r} and {b!r} must be RpmVersion or strings")

    a_epoch = a.epoch or "0"
    b_epoch = b.epoch or "0"

    # First compare the epoch, if set.  If the epoch's are not the same, then
    # the higher one wins no matter what the rest of the EVR is.
    if a_epoch != b_epoch:
        epoch_compare = _compare_version_strings(a_epoch, b_epoch)
        if epoch_compare != 0:
            return epoch_compare  # a > b

    # Epoch is the same, if version + release are the same we have a match
    if (a.version == b.version) and (a.release == b.release):
        return 0  # a == b

    # Compare version first, if version is equal then compare release
    version_compare = _compare_version_strings(a.version, b.version)
    if version_compare != 0:  # a > b || a < b
        return version_compare

    return _compare_version_strings(a.release, b.release)


# internal use: each individual component of the EVR is compared using this function
def _compare_version_strings(first, second):
    first = first.encode("utf-8")
    second = second.encode("utf-8")

    if first == second:
        return 0

    def not_alphanumeric_tilde_or_caret(c):
        return not (
            (ord(b"a") <= c <= ord(b"z"))
            or (ord(b"A") <= c <= ord(b"Z"))
            or (ord(b"0") <= c <= ord(b"9"))
            or c == ord(b"~")
            or c == ord(b"^")
        )

    def trim_start_matches(data, predicate):
        """Trim leading bytes that match the predicate"""
        start = 0
        while start < len(data) and predicate(data[start]):
            start += 1
        return data[start:]

    def strip_prefix(data, prefix):
        """Strip prefix from data, return (stripped_data, was_stripped)"""
        if data.startswith(prefix):
            return data[len(prefix) :], True
        return data, False

    def matching_contiguous(data, predicate):
        """Match contiguous characters that satisfy predicate"""
        if not data:
            return None, data

        if not predicate(data[0]):
            return None, data

        end = 0
        while end < len(data) and predicate(data[end]):
            end += 1

        return data[:end], data[end:]

    version1_part = first
    version2_part = second

    while True:
        # Strip any leading non-alphanumeric, non-tilde, non-caret characters
        version1_part = trim_start_matches(version1_part, not_alphanumeric_tilde_or_caret)
        version2_part = trim_start_matches(version2_part, not_alphanumeric_tilde_or_caret)

        # Tilde separator parses as "older" or lesser version
        version1_stripped, version1_had_tilde = strip_prefix(version1_part, b"~")
        version2_stripped, version2_had_tilde = strip_prefix(version2_part, b"~")

        if version1_had_tilde and not version2_had_tilde:
            return -1
        elif not version1_had_tilde and version2_had_tilde:
            return 1
        elif version1_had_tilde and version2_had_tilde:
            version1_part = version1_stripped
            version2_part = version2_stripped
            continue

        # Caret means the version is less... Unless the other version
        # has ended, then do the exact opposite.
        version1_stripped, version1_had_caret = strip_prefix(version1_part, b"^")
        version2_stripped, version2_had_caret = strip_prefix(version2_part, b"^")

        if version1_had_caret and not version2_had_caret:
            if not version2_part:  # second has ended
                return 1  # first > second
            else:  # second continues
                return -1  # first < second
        elif not version1_had_caret and version2_had_caret:
            if not version1_part:  # first has ended
                return -1  # first < second
            else:  # first continues
                return 1  # first > second
        elif version1_had_caret and version2_had_caret:
            version1_part = version1_stripped
            version2_part = version2_stripped
            continue

        # Check if we've run out of characters
        if not version1_part and not version2_part:
            return 0
        elif not version1_part:
            return -1
        elif not version2_part:
            return 1

        # Parse numeric or alphabetic segments
        def is_digit(c):
            return ord(b"0") <= c <= ord(b"9")

        def is_alpha(c):
            return (ord(b"a") <= c <= ord(b"z")) or (ord(b"A") <= c <= ord(b"Z"))

        if version1_part and is_digit(version1_part[0]):
            # First starts with digit - extract numeric segment
            segment1, version1_part = matching_contiguous(version1_part, is_digit)

            if version2_part and is_digit(version2_part[0]):
                # Both numeric
                segment2, version2_part = matching_contiguous(version2_part, is_digit)

                # Strip leading zeros
                segment1 = segment1.lstrip(b"0")
                segment2 = segment2.lstrip(b"0")

                # Compare by length first (more digits = larger number)
                if len(segment1) < len(segment2):
                    return -1
                elif len(segment1) > len(segment2):
                    return 1
                else:
                    # Same length, compare lexicographically
                    if segment1 < segment2:
                        return -1
                    elif segment1 > segment2:
                        return 1
                    # Equal, continue to next segment
            else:
                # First is numeric, second is not - numeric wins
                return 1
        else:
            # First starts with alpha or we're at end
            if version1_part:
                segment1, version1_part = matching_contiguous(version1_part, is_alpha)
            else:
                segment1 = b""

            if version2_part and is_digit(version2_part[0]):
                # First is alpha, second is numeric - numeric wins
                return -1
            else:
                # Both alpha or at least one is empty
                if version2_part:
                    segment2, version2_part = matching_contiguous(version2_part, is_alpha)
                else:
                    segment2 = b""

                # Compare alphabetically
                if segment1 < segment2:
                    return -1
                elif segment1 > segment2:
                    return 1
                # Equal, continue to next segment

    # Should not reach here due to the checks above, but just in case
    raise RuntimeError("somehow escaped the loop during version comparison")



import logging
import os
import shutil
import tempfile
from gettext import gettext as _
from typing import NamedTuple
from uuid import UUID

import createrepo_c as cr
import libcomps
from django.conf import settings
from django.core.files import File
from django.db.models import Q
from pulpcore.plugin.models import (
    AsciiArmoredDetachedSigningService,
    ContentArtifact,
    ProgressReport,
    PublishedArtifact,
    PublishedMetadata,
    RepositoryContent,
    RepositoryVersion,
)

from pulp_rpm.app.comps import dict_to_strdict
from pulp_rpm.app.constants import (
    ALLOWED_CHECKSUM_ERROR_MSG,
    CHECKSUM_TYPES,
    COMPRESSION_TYPES,
    LAYOUT_TYPES,
    PACKAGES_DIRECTORY,
)
from pulp_rpm.app.kickstart.treeinfo import PulpTreeInfo, TreeinfoData
from pulp_rpm.app.models import (
    DistributionTree,
    Modulemd,
    ModulemdDefaults,
    ModulemdObsolete,
    Package,
    PackageCategory,
    PackageEnvironment,
    PackageGroup,
    PackageLangpacks,
    RepoMetadataFile,
    RpmPublication,
    UpdateRecord,
)
from pulp_rpm.app.serializers import RpmPublicationSerializer
from pulp_rpm.app.shared_utils import format_nevra

log = logging.getLogger(__name__)

REPODATA_PATH = "repodata"

# lift dynaconf lookups outside of loops
ALLOWED_CONTENT_CHECKSUMS = settings.ALLOWED_CONTENT_CHECKSUMS
RPM_METADATA_USE_REPO_PACKAGE_TIME = settings.RPM_METADATA_USE_REPO_PACKAGE_TIME


class PackageInfo(NamedTuple):
    """
    Data about a package being published that needs to be shared with the repo metadata.

    Attributes:
        caid (UUID): ContentArtifact ID.
        path (str): The relative URL where the package will be published.
        checksum_type (str): The type (eg. sha256) of the checksum.
        checksum (str): The checksum value of the package.

    """

    caid: UUID
    path: str
    checksum_type: str
    checksum: str


class PkgBuild(NamedTuple):
    """
    Data about a package build for collision resolution.

    Attributes:
        cid (UUID): Content ID.
        epoch (int): The Epoch of the package build.
        build_time (int): The build time (unix timestamp format) of the package.

    """

    cid: UUID
    epoch: int
    build_time: int


class _CollisionManager:
    """Helper to collect the "winning" packages when there are collisions on NEVRA or URL path."""

    def __init__(self) -> None:
        self.cid_to_second_path: dict[UUID, str] = {}
        self._nevra_to_pkg: dict[str, PkgBuild] = {}
        self._path_to_pkg: dict[str, PkgBuild] = {}
        self._second_path_to_pkg: dict[str, PkgBuild] = {}
        self._banned_cids: set[UUID] = set()

    def _pkg_to_ignore(
        self, new: PkgBuild, index_to_pkg: dict[str, PkgBuild], index: str
    ) -> PkgBuild | None:
        old = index_to_pkg.get(index, None)
        if not old or old.cid in self._banned_cids:
            return None

        log.warning(
            _(
                "Duplicate packages found competing for {index}, selected the one with "
                "the most recent epoch or build time."
            ).format(index=index)
        )
        if old.epoch > new.epoch or (old.epoch == new.epoch and old.build_time >= new.build_time):
            return new
        return old

    def add(self, pkg: PkgBuild, nevra: str, path: str, second_path: str | None) -> None:
        """
        Add a package build to the collision manager. Ignore it if it is "worse" than an existing
        package that collides on nevra or path. If it's "better" and collides, mark the older one as
        ignored so it will not be included when `retained_cids` is called.

        Args:
            pkg (PkgBuild): Information about the package build we're adding.
            nevra (str): NEVRA of the package. Checked for collisions in repo metadata.
            path (str): URL path of the package. Checked for collisions in published artifact URLs.
            second_path (str | None):
                An optional secondary URL path of the package. Only will exist in the nested_by_both
                case. Checked for collisions like `path`, except collisions here only prevent the
                "worse" package from being published at this secondary path, not from being
                published at in general.
        """
        # If there is a collision on nevra or path, ignore the older package
        to_ignore_nevra = self._pkg_to_ignore(pkg, self._nevra_to_pkg, nevra)
        to_ignore_path = self._pkg_to_ignore(pkg, self._path_to_pkg, path)

        if to_ignore_nevra == pkg or to_ignore_path == pkg:
            return  # Just ignore it before it's added anywhere.

        # Else remember it.
        self._nevra_to_pkg[nevra] = pkg
        self._path_to_pkg[path] = pkg

        # We have record cids that were once added but later ignored; we don't know all its keys.
        if to_ignore_nevra:
            self._banned_cids.add(to_ignore_nevra.cid)
            self.cid_to_second_path.pop(to_ignore_nevra.cid, None)
        if to_ignore_path:
            self._banned_cids.add(to_ignore_path.cid)
            self.cid_to_second_path.pop(to_ignore_path.cid, None)

        # In the nested_by_both case we may have a package that should be published in general, but
        # collides on the alphabetical path and should be ignored for that path only.
        if second_path:
            to_ignore = self._pkg_to_ignore(pkg, self._second_path_to_pkg, second_path)
            if to_ignore == pkg:
                return
            elif to_ignore:
                self.cid_to_second_path.pop(to_ignore.cid, None)

            self._second_path_to_pkg[second_path] = pkg
            self.cid_to_second_path[pkg.cid] = second_path

    def retained_cids(self) -> list[UUID]:
        """
        Return a list of retained Content IDs, excluding those that collide on NEVRA or Path.

        Returns:
            list: List of Content IDs that should be published.
        """
        return [pkg.cid for pkg in self._nevra_to_pkg.values() if pkg.cid not in self._banned_cids]


class PublicationData:
    """
    Encapsulates data relative to publication.

    Attributes:
        publication (pulpcore.plugin.models.Publication): A Publication to populate.
        sub_repos (list): A list of tuples with sub_repos data.
        repomdrecords (list): A list of tuples with repomdrecords data.

    """

    def __init__(self, publication, checksum_types):
        """
        Setting Publication data.

        Args:
            publication (pulpcore.plugin.models.Publication): A Publication to populate.

        """
        self.publication = publication
        self.sub_repos = []
        self.repomdrecords = []
        self.checksum_types = checksum_types
        self.packages: dict[UUID, PackageInfo] = {}

    def prepare_metadata_files(self, content, folder=None):
        """
        Copies metadata files from the Artifact storage.

        Args:
            content (pulpcore.plugin.models.Content): content set.

        Keyword Args:
            folder(str): name of the directory.

        Returns:
            repomdrecords (list): A list of tuples with repomdrecords data.

        """
        repomdrecords = []
        repo_metadata_files = RepoMetadataFile.objects.filter(pk__in=content).prefetch_related(
            "contentartifact_set"
        )

        for repo_metadata_file in repo_metadata_files:
            if repo_metadata_file.unsupported_metadata_type:
                # Normally these types are not synced in the first place, we skip them here, since
                # they might still exist in old repo versions from before we started excluding them.
                continue
            content_artifact = repo_metadata_file.contentartifact_set.get()
            current_file = content_artifact.artifact.file.file
            path = content_artifact.relative_path.split("/")[-1]
            if repo_metadata_file.checksum in path:
                # filenames can be checksum-xxxx.yyy - but can also be checksum-mmm-nnn-ooo.yyy
                # Some old repos might have filename of checksum only so we need to take care
                # of that too
                # split off the checksum, keep the rest
                filename = path.split("-")[1:]
                if filename:
                    path = "-".join(filename)
            if folder:
                path = os.path.join(folder, path)
            with open(path, "wb") as new_file:
                shutil.copyfileobj(current_file, new_file)
                repomdrecords.append((repo_metadata_file.data_type, new_file.name))

        return repomdrecords

    def publish_artifacts(self, content, prefix=""):
        """
        Create PublishedArtifacts for each Artifact. Special considerations for Packages:

        1. Respect the layout of the publication.
           1. "flat": "Packages/xxx.rpm"
           2. "nested_alphabetically": "Packages/f/filename.rpm"   - The "canonical" repo format.
           3. "nested_by_digest": "Packages/f/by-digest/xxxxxx-filename.rpm"
           4. "nested_by_both": Both 2 and 3. Each rpm will result in 2 PublishedArtifacts.

        2. Handle "duplicate" packages.

           There are two ways that different packages (different checksums) can "collide":
              a) Packages with the same NEVRA (Name, Epoch, Version, Release, Arch) will
                 collide in the repo metadata. One of them will "win" in the client. This can be the
                 case if a signed and unsigned package are both published in the same repo for
                 example, but also if people just reuse NEVRAs for different builds.
              b) Packages that result in the same URL will collide at the http level. This can be
                 different from case A if they are not using the NEVRA format for relative_path.
                 They may not have anywhere close to the same NEVRA, just the same filename.

           In both cases we have to pick one - which is essentially what the rest of the RPM
           Ecosystem does when faced with the impossible. This code takes the one with the
           most recent build time which is the same heuristic used by Yum/DNF/Zypper. The "winner"
           package must be consistent here and when writing the repo metadata later.

        3. Use the correct hash for artifacts.

           We want to support publishing with a different checksum type than the one built-in to
           the package itself, so we need to get the correct checksums somehow if there is an
           override. We must also take into consideration that if the package has not been
           downloaded the only checksum that is available is the one built-in.

           Since this lookup goes from Package->Content->ContentArtifact->Artifact, performance is
           a challenge. We use ContentArtifact as our starting point because it enables us to work
           with simple foreign keys and avoid messing with the many-to-many relationship, which
           doesn't work with select_related() and performs poorly with prefetch_related(). This is
           fine because we know that Packages should only ever have one artifact per content.

        All three of these considerations are relevant both here when creating the
        PublishedArtifacts, as well as later when generating the repo metadata that references them.
        Do these things once here and record the info for use later to avoid lots of redundant
        lookups, as well as guarantee we're making the same decisions in both places.

        Note that this only impacts user-created publications, which produce the "standard"
        RPM layout of repo/Packages/f/foo.rpm. A publication created by mirror-sync retains
        whatever layout their "upstream" repo-metadata dictates.

        Args:
            content (pulpcore.plugin.models.Content): content set.
            prefix (str): a relative path prefix for the published artifact

        Returns:
            dict: Mapping of content_id to PackageInfo for retained packages.
        """

        def nested_alphabetically_path(pkg_filename):
            """Returns the path to use for nested_alphabetically. Used twice, define once."""
            return os.path.join(PACKAGES_DIRECTORY, pkg_filename[0].lower(), pkg_filename)

        def nested_by_digest_path(pkg_filename, checksum):
            """Returns the path to use for nested_by_digest. Used twice, define once."""
            # Regardless of checksum type, let's use the first 6 characters of the checksum
            # to create a nested directory structure.
            if len(checksum) < 6:
                raise ValueError(
                    f"Checksum {checksum} is unknown or too short to use for "
                    f"{self.publication.layout} publishing."
                )
            return os.path.join(
                PACKAGES_DIRECTORY,
                pkg_filename[0].lower(),
                "by-digest",
                f"{checksum[:6]}-{pkg_filename}",
            )

        def flat_path(pkg_filename):
            """Returns the path to use for flat layout. Define to keep it close to the others."""
            return os.path.join(PACKAGES_DIRECTORY, pkg_filename)

        published_artifacts = []
        requested_checksum_type = get_checksum_type(self.checksum_types)
        layout = self.publication.layout
        collision_manager = _CollisionManager()
        cid_to_pkginfo: dict[UUID, PackageInfo] = {}

        # Special Handling for Packages first
        contentartifact_qs = ContentArtifact.objects.filter(content__in=content).filter(
            content__pulp_type=Package.get_pulp_type()
        )

        fields = [
            "pk",
            "content_id",
            "relative_path",
            "content__rpm_package__checksum_type",
            "content__rpm_package__pkgId",
            "content__rpm_package__name",
            "content__rpm_package__epoch",
            "content__rpm_package__version",
            "content__rpm_package__release",
            "content__rpm_package__arch",
            "content__rpm_package__time_build",
        ]
        artifact_checksum = None
        if requested_checksum_type:
            artifact_checksum = f"artifact__{requested_checksum_type}"
            fields.append(artifact_checksum)

        for row in contentartifact_qs.values(*fields).iterator():
            # content_id is the same as the Package PK, which is used later when generating repo
            # metadata. The contentartifact PK is different, and is used here for PublishedArtifact.
            # There is no such thing as a multi-Artifact RPM Package, so in practice these are 1:1,
            # even though that's not enforced at the DB level. (There _are_ multi-Artifact Contents
            # in general, such as deb src packages - where one logical Content maps to three actual
            # files, and a unique file may appear in any number of deb src packages - so pulpcore
            # allows a many-to-many relationship. But this is not applicable to RPM Packages.)

            # First, get some basic package data
            caid = row["pk"]
            cid = row["content_id"]
            build_time = row["content__rpm_package__time_build"]
            epoch = row["content__rpm_package__epoch"]
            nevra = format_nevra(
                row["content__rpm_package__name"],
                epoch,
                row["content__rpm_package__version"],
                row["content__rpm_package__release"],
                row["content__rpm_package__arch"],
            )
            epoch = int(epoch) if epoch else 0  # epoch is always an int if defined
            pkg_build = PkgBuild(cid=cid, epoch=epoch, build_time=build_time)

            # Second, get the checksum / checksum type, defaulting to rpm_package__pkgId if needed
            if requested_checksum_type:
                checksum = row.get(artifact_checksum, None)
                checksum_type = requested_checksum_type

            if not requested_checksum_type or not checksum:
                checksum = row["content__rpm_package__pkgId"]
                checksum_type = row["content__rpm_package__checksum_type"]
                if checksum_type not in ALLOWED_CONTENT_CHECKSUMS:
                    raise ValueError(
                        "Package with pkgId {} as content unit {} contains forbidden checksum type "
                        "'{}', thus can't be published. {}".format(
                            checksum,
                            cid,
                            checksum_type,
                            ALLOWED_CHECKSUM_ERROR_MSG,
                        )
                    )

            # Third, compute the path based on layout
            pkg_filename = os.path.basename(row["relative_path"])
            second_path = None
            if layout == LAYOUT_TYPES.NESTED_ALPHABETICALLY:
                path = nested_alphabetically_path(pkg_filename)
            elif layout == LAYOUT_TYPES.FLAT:
                path = flat_path(pkg_filename)
            elif layout == LAYOUT_TYPES.NESTED_BY_DIGEST:
                path = nested_by_digest_path(pkg_filename, checksum)
            elif layout == LAYOUT_TYPES.NESTED_BY_BOTH:
                path = nested_by_digest_path(pkg_filename, checksum)
                second_path = nested_alphabetically_path(pkg_filename)
            else:
                raise ValueError(f"Layout value {layout} is unsupported by this version")

            pkg_info = PackageInfo(
                caid=caid, path=path, checksum_type=checksum_type, checksum=checksum
            )
            collision_manager.add(pkg_build, nevra, path, second_path)
            cid_to_pkginfo[cid] = pkg_info

        # Filter cid_to_pkginfo to only the retained packages
        retained_cids = collision_manager.retained_cids()
        cid_to_pkginfo = {k: cid_to_pkginfo[k] for k in retained_cids}

        # Finally create the PublishedArtifacts for the remaining packages
        for cid, pkg_info in cid_to_pkginfo.items():
            published_artifacts.append(
                PublishedArtifact(
                    relative_path=os.path.join(prefix, pkg_info.path),
                    publication=self.publication,
                    content_artifact_id=pkg_info.caid,
                )
            )
            if second_path := collision_manager.cid_to_second_path.get(cid, None):
                # also add the nested_alphabetically path
                published_artifacts.append(
                    PublishedArtifact(
                        relative_path=os.path.join(prefix, second_path),
                        publication=self.publication,
                        content_artifact_id=pkg_info.caid,
                    )
                )

        # Handle the non-packages
        is_treeinfo = Q(relative_path__in=["treeinfo", ".treeinfo"])
        unpublishable_types = Q(
            content__pulp_type__in=[
                RepoMetadataFile.get_pulp_type(),
                Modulemd.get_pulp_type(),
                ModulemdDefaults.get_pulp_type(),
                # already dealt with
                Package.get_pulp_type(),
            ]
        )

        contentartifact_qs = (
            ContentArtifact.objects.filter(content__in=content)
            .exclude(unpublishable_types)
            .exclude(is_treeinfo)
        )

        for content_artifact in contentartifact_qs.values("pk", "relative_path").iterator():
            published_artifacts.append(
                PublishedArtifact(
                    relative_path=content_artifact["relative_path"],
                    publication=self.publication,
                    content_artifact_id=content_artifact["pk"],
                )
            )

        PublishedArtifact.objects.bulk_create(published_artifacts, batch_size=2000)
        return cid_to_pkginfo

    def handle_sub_repos(self, distribution_tree):
        """
        Get sub-repo content and publish them.

        Args:
            distribution_tree (pulp_rpm.models.DistributionTree): A distribution_tree object.

        """
        original_treeinfo_content_artifact = distribution_tree.contentartifact_set.get(
            relative_path__in=[".treeinfo", "treeinfo"]
        )
        orig_artifact = original_treeinfo_content_artifact.artifact
        artifact_file = orig_artifact.pulp_domain.get_storage().open(orig_artifact.file.name)
        with tempfile.NamedTemporaryFile("wb", dir=".") as temp_file:
            shutil.copyfileobj(artifact_file, temp_file)
            temp_file.flush()
            treeinfo = PulpTreeInfo()
            treeinfo.load(f=temp_file.name)
            treeinfo_data = TreeinfoData(treeinfo.parsed_sections())

            # rewrite the treeinfo file such that the variant repository and package location
            # is a relative subtree
            treeinfo.rewrite_subrepo_paths(treeinfo_data)

            # TODO: better way to do this?
            main_variant = treeinfo.original_parser._sections.get("general", {}).get(
                "variant", None
            )
            treeinfo_file = tempfile.NamedTemporaryFile(dir=".")
            treeinfo.dump(treeinfo_file.name, main_variant=main_variant)
            with open(treeinfo_file.name, "rb") as treeinfo_fd:
                PublishedMetadata.create_from_file(
                    relative_path=original_treeinfo_content_artifact.relative_path,
                    publication=self.publication,
                    file=File(treeinfo_fd),
                )
        artifact_file.close()
        relations = ["addon", "variant"]
        for relation in relations:
            addons_or_variants = getattr(distribution_tree, f"{relation}s").all()
            for addon_or_variant in addons_or_variants:
                if not addon_or_variant.repository:
                    # a variant of the main repo
                    continue
                repository = addon_or_variant.repository.cast()
                repository_version = repository.latest_version()

                if repository_version and repository.user_hidden:
                    addon_or_variant_id = getattr(addon_or_variant, f"{relation}_id")
                    self.sub_repos.append(
                        (
                            addon_or_variant_id,
                            repository_version.content,
                        )
                    )

    def populate(self):
        """
        Populate a publication.

        Create published artifacts for a publication.

        """
        main_content = self.publication.repository_version.content
        self.repomdrecords = self.prepare_metadata_files(main_content)

        self.packages = self.publish_artifacts(main_content)

        distribution_trees = DistributionTree.objects.filter(pk__in=main_content).prefetch_related(
            "addons",
            "variants",
            "addons__repository",
            "variants__repository",
            "contentartifact_set",
        )

        for distribution_tree in distribution_trees:
            self.handle_sub_repos(distribution_tree)

        for name, content in self.sub_repos:
            os.mkdir(name)
            setattr(self, f"{name}_content", content)
            setattr(self, f"{name}_checksums", self.checksum_types)
            setattr(self, f"{name}_repomdrecords", self.prepare_metadata_files(content, name))
            setattr(self, f"{name}_packages", self.publish_artifacts(content, prefix=name))


def get_checksum_type(checksum_types, default=CHECKSUM_TYPES.SHA256):
    """
    Get checksum algorithm for publishing metadata.

    Args:
        checksum_types (dict): Checksum types for metadata and packages.
    Kwargs:
        default: The checksum type used if there is no specified nor original checksum type.
    """
    general = checksum_types.get("general")
    metadata = checksum_types.get("metadata")
    # fallback order
    checksum_type = general or metadata or default
    # "sha" -> "SHA" -> "CHECKSUM_TYPES.SHA" -> "sha1"
    normalized_checksum_type = getattr(CHECKSUM_TYPES, checksum_type.upper())
    return normalized_checksum_type


def cr_checksum_type_from_string(checksum_type):
    """
    Convert checksum type from string to createrepo_c enum variant.
    """
    # "sha1" -> "SHA1" -> "cr.SHA1"
    return getattr(cr, checksum_type.upper())


def publish(
    repository_version_pk,
    metadata_signing_service=None,
    checkpoint=False,
    checksum_types=None,
    checksum_type=None,
    repo_config=None,
    compression_type=COMPRESSION_TYPES.GZ,
    layout=None,
    *args,
    **kwargs,
):
    """
    Create a Publication based on a RepositoryVersion.

    Args:
        repository_version_pk (str): Create a publication from this repository version.
        metadata_signing_service (pulpcore.app.models.AsciiArmoredDetachedSigningService):
            A reference to an associated signing service.
        checkpoint (bool): Whether to create a checkpoint publication.
        checksum_types (dict): Checksum types for metadata and packages.
        repo_config (JSON): repo config that will be served by distribution
        compression_type(pulp_rpm.app.constants.COMPRESSION_TYPES):
            Compression type to use for metadata files.
        layout(pulp_rpm.app.constants.LAYOUT_TYPES):
            How to layout the package files within the publication (flat, nested, etc.)

    """
    repository_version = RepositoryVersion.objects.get(pk=repository_version_pk)
    repository = repository_version.repository.cast()
    checksum_types = checksum_types or {}
    # currently unused, but prep for eliminating "checksum_types due to zero-downtime requirements"
    if checksum_type:
        checksum_types = {"general": checksum_type}

    if metadata_signing_service:
        metadata_signing_service = AsciiArmoredDetachedSigningService.objects.get(
            pk=metadata_signing_service
        )

    if layout is None:
        # Can't simply set a default in the function signature because some code calls this function
        # with an explicit None value.
        layout = LAYOUT_TYPES.NESTED_ALPHABETICALLY

    log.info(
        _("Publishing: repository={repo}, version={version}").format(
            repo=repository.name,
            version=repository_version.number,
        )
    )
    with tempfile.TemporaryDirectory(dir="."):
        with RpmPublication.create(repository_version, checkpoint=checkpoint) as publication:
            checksum_type = get_checksum_type(checksum_types)
            publication.checksum_type = checksum_type
            publication.compression_type = compression_type
            publication.layout = layout
            publication.repo_config = repo_config

            publication_data = PublicationData(publication, checksum_types)
            publication_data.populate()

            total_repos = 1 + len(publication_data.sub_repos)
            pb_data = dict(
                message="Generating repository metadata",
                code="publish.generating_metadata",
                total=total_repos,
            )
            with ProgressReport(**pb_data) as publish_pb:
                content = publication.repository_version.content

                # Main repo
                generate_repo_metadata(
                    content,
                    publication,
                    checksum_types,
                    publication_data.repomdrecords,
                    metadata_signing_service=metadata_signing_service,
                    compression_type=compression_type,
                    retained_packages=publication_data.packages,
                )
                publish_pb.increment()

                for sub_repo in publication_data.sub_repos:
                    name = sub_repo[0]
                    content = getattr(publication_data, f"{name}_content")
                    extra_repomdrecords = getattr(publication_data, f"{name}_repomdrecords")
                    packages = getattr(publication_data, f"{name}_packages")
                    generate_repo_metadata(
                        content,
                        publication,
                        checksum_types,
                        extra_repomdrecords,
                        name,
                        metadata_signing_service=metadata_signing_service,
                        compression_type=compression_type,
                        retained_packages=packages,
                    )
                    publish_pb.increment()

            log.info(_("Publication: {publication} created").format(publication=publication.pk))
            serialized_pub = RpmPublicationSerializer(
                instance=publication, context={"request": None}
            ).data
            return serialized_pub


def generate_repo_metadata(
    content,
    publication,
    checksum_types,
    extra_repomdrecords,
    sub_folder=None,
    metadata_signing_service=None,
    compression_type=COMPRESSION_TYPES.GZ,
    retained_packages: dict[UUID, PackageInfo] = {},
):
    """
    Creates a repomd.xml file.

    Args:
        content(app.models.Content): A DB Content set of all original artifacts in the publication.
        publication(pulpcore.plugin.models.Publication): the publication
        extra_repomdrecords(list): list with data relative to repo metadata files
        sub_folder(str): name of the folder for sub repos
        metadata_signing_service (pulpcore.app.models.AsciiArmoredDetachedSigningService):
            A reference to an associated signing service.
        compression_type(pulp_rpm.app.constants.COMPRESSION_TYPES):
            Compression type to use for metadata files.
        retained_packages(dict):
            A dictionary of content_id to PackageInfo for packages that should actually be included
            in the repository metadata. Will be used to filter `content` and add additional info.

    """
    cwd = os.getcwd()
    repodata_path = REPODATA_PATH
    has_modules = False
    has_comps = False
    requested_checksum_type = get_checksum_type(checksum_types)

    if requested_checksum_type not in ALLOWED_CONTENT_CHECKSUMS:
        raise ValueError(
            "Disallowed checksum type '{}' was requested to be used for publication: {}".format(
                requested_checksum_type, ALLOWED_CHECKSUM_ERROR_MSG
            )
        )

    if sub_folder:
        cwd = os.path.join(cwd, sub_folder)
        repodata_path = os.path.join(sub_folder, repodata_path)

    # Prepare metadata files
    if compression_type == COMPRESSION_TYPES.ZSTD:
        cr_compression_type = cr.ZSTD
    elif compression_type == COMPRESSION_TYPES.NONE:
        cr_compression_type = cr.NO_COMPRESSION
    else:
        # Gzip is the default option & fallback should the value be something unexpected
        cr_compression_type = cr.GZ
    total_packages = len(retained_packages)

    if RPM_METADATA_USE_REPO_PACKAGE_TIME:
        # gather the times the packages were added to the repo
        repo_content = (
            RepositoryContent.objects.filter(
                repository=publication.repository,
                version_added__number__lte=publication.repository_version.number,
            )
            .exclude(version_removed__number__lte=publication.repository_version.number)
            .values_list("content", "pulp_created")
        )
        repo_pkg_times = {pk: created.timestamp() for pk, created in repo_content}

    repomd_path = os.path.join(repodata_path, "repomd.xml")
    mod_yml_path = os.path.join(repodata_path, "modules.yaml")
    comps_xml_path = os.path.join(repodata_path, "comps.xml")

    cr_checksum_type = cr_checksum_type_from_string(publication.checksum_type)

    # Process all packages
    with cr.RepositoryWriter(
        cwd, compression=cr_compression_type, checksum_type=cr_checksum_type
    ) as writer:
        writer.set_num_of_pkgs(total_packages)

        # If the repository is empty, use a revision of 0
        # See: https://pulp.plan.io/issues/9402
        if not content.exists():
            writer.repomd.revision = "0"
        for package in (
            Package.objects.filter(pk__in=content).order_by("name", "evr").iterator(chunk_size=200)
        ):
            if package.pk not in retained_packages:
                continue
            pkg = package.to_createrepo_c()

            # rewrite these fields with the desired ones
            retained_pkg_info = retained_packages[package.pk]
            pkg.checksum_type = retained_pkg_info.checksum_type
            pkg.pkgId = retained_pkg_info.checksum
            pkg.location_href = retained_pkg_info.path

            if RPM_METADATA_USE_REPO_PACKAGE_TIME:
                pkg.time_file = repo_pkg_times[package.pk]

            writer.add_pkg(pkg)

        # Process update records
        update_records = UpdateRecord.objects.filter(pk__in=content).order_by("id", "digest")
        for update_record in update_records.iterator():
            writer.add_update_record(update_record.to_createrepo_c())

        # Process modulemd, modulemd_defaults and obsoletes
        with open(mod_yml_path, "ab") as mod_yml:
            modulemds = Modulemd.objects.filter(pk__in=content).order_by(
                *Modulemd.natural_key_fields()
            )
            for modulemd in modulemds.iterator():
                mod_yml.write(modulemd.snippet.encode())
                mod_yml.write(b"\n")
                has_modules = True
            modulemd_defaults = ModulemdDefaults.objects.filter(pk__in=content).order_by(
                *ModulemdDefaults.natural_key_fields()
            )
            for default in modulemd_defaults.iterator():
                mod_yml.write(default.snippet.encode())
                mod_yml.write(b"\n")
                has_modules = True
            modulemd_obsoletes = ModulemdObsolete.objects.filter(pk__in=content).order_by(
                *ModulemdObsolete.natural_key_fields()
            )
            for obsolete in modulemd_obsoletes.iterator():
                mod_yml.write(obsolete.snippet.encode())
                mod_yml.write(b"\n")
                has_modules = True

        # Process comps
        comps = libcomps.Comps()
        for pkg_grp in PackageGroup.objects.filter(pk__in=content).order_by("id").iterator():
            group = pkg_grp.pkg_grp_to_libcomps()
            comps.groups.append(group)
            has_comps = True
        for pkg_cat in PackageCategory.objects.filter(pk__in=content).order_by("id").iterator():
            cat = pkg_cat.pkg_cat_to_libcomps()
            comps.categories.append(cat)
            has_comps = True
        for pkg_env in PackageEnvironment.objects.filter(pk__in=content).order_by("id").iterator():
            env = pkg_env.pkg_env_to_libcomps()
            comps.environments.append(env)
            has_comps = True
        package_langpacks = PackageLangpacks.objects.filter(pk__in=content).order_by(
            *PackageLangpacks.natural_key_fields()
        )
        for pkg_lng in package_langpacks.iterator():
            comps.langpacks = dict_to_strdict(pkg_lng.matches)
            has_comps = True

        comps.toxml_f(
            comps_xml_path,
            xml_options={
                "default_explicit": True,
                "empty_groups": True,
                "empty_packages": True,
                "uservisible_explicit": True,
            },
        )

        if has_modules:
            writer.add_repomd_metadata("modules", mod_yml_path, use_compression=False)

        if has_comps:
            writer.add_repomd_metadata("group", comps_xml_path, use_compression=False)

        for name, record in extra_repomdrecords:
            writer.add_repomd_metadata(name, record)

    for record in writer.repomd.records:
        path = os.path.join(repodata_path, os.path.basename(record.location_href))
        with open(path, "rb") as repodata_fd:
            PublishedMetadata.create_from_file(
                relative_path=path,
                publication=publication,
                file=File(repodata_fd),
            )

    if metadata_signing_service:
        signing_service = AsciiArmoredDetachedSigningService.objects.get(
            pk=metadata_signing_service
        )
        sign_results = signing_service.sign(repomd_path)

        # https://github.com/pulp/pulp_rpm/issues/3526
        signature_file_path = sign_results["signature"]
        if os.stat(signature_file_path).st_size == 0:
            log.error(f"{signature_file_path} is 0 bytes! sign_results: {sign_results}")
            raise Exception("Signature file is 0 bytes")

        # publish a signed file
        with open(sign_results["file"], "rb") as signed_file_fd:
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(repodata_path, os.path.basename(sign_results["file"])),
                publication=publication,
                file=File(signed_file_fd),
            )

        # publish a detached signature
        with open(sign_results["signature"], "rb") as signature_fd:
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(
                    repodata_path, os.path.basename(sign_results["signature"])
                ),
                publication=publication,
                file=File(signature_fd),
            )

        # publish a public key required for further verification
        pubkey_name = "repomd.xml.key"
        with open(pubkey_name, "wb+") as f:
            f.write(signing_service.public_key.encode("utf-8"))
            f.flush()
            # important! as the file has already been opened and used, it will be treated as a
            # cursor and when calculating the checksum it will calculate the checksum of nothing.
            f.seek(0)
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(repodata_path, pubkey_name),
                publication=publication,
                file=File(f),
            )
    else:
        with open(repomd_path, "rb") as repomd_fd:
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(repodata_path, os.path.basename(repomd_path)),
                publication=publication,
                file=File(repomd_fd),
            )



# Generated by Django 4.2.27 on 2026-01-06 16:48

from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import pulpcore.app.models.base


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0066_layout_choices'),
    ]

    operations = [
        migrations.CreateModel(
            name='RpmPackageSigningResult',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('original_package_sha256', models.TextField(max_length=64)),
                ('package_signing_fingerprint', models.TextField(max_length=40)),
                ('result_package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.content')),
            ],
            options={
                'unique_together': {('original_package_sha256', 'package_signing_fingerprint')},
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]



import shutil
import subprocess
import tempfile
import typing as t
from hashlib import sha256
from pathlib import Path
from collections import defaultdict

import createrepo_c as cr
from django.conf import settings
from django.utils.dateparse import parse_datetime
from importlib_resources import files
from pulpcore.plugin.exceptions import InvalidSignatureError
from pulp_rpm.app.rpm_version import RpmVersion


def annotate_with_age(qs):
    """Provide an "age" score for each Package object in the queryset.

    Annotate the Package objects with an "age". Age is calculated by partitioning the
    Packages by name and architecture and ordering the packages in each group by 'evr',
    which is the relative "age" within the group. The newest package gets age=1, second
    newest age=2, and so on.

    A second partition by architecture is important because there can be packages with
    the same name and version numbers but they are not interchangeable because they have
    differing arch, such as 'x86_64' and 'i686', or 'src' (SRPM) and any other arch.
    """
    # Get packages in current queryset with their basic info
    packages = list(qs.values("pk", "name", "arch", "epoch", "version", "release"))

    # Group packages by name and arch
    groups = defaultdict(list)
    for pkg in packages:
        key = (pkg["name"], pkg["arch"])
        groups[key].append(pkg)

    # Calculate age for each group
    age_mapping = {}
    for group_packages in groups.values():
        # Sort by EVR (newest first)
        group_packages.sort(
            key=lambda p: RpmVersion(p["epoch"], p["version"], p["release"]), reverse=True
        )

        # Assign ages (1 = newest, 2 = second newest, etc.)
        for age, pkg in enumerate(group_packages, 1):
            age_mapping[pkg["pk"]] = age

    # Create a queryset with age annotation
    # We'll use a CASE statement to map PKs to ages
    from django.db.models import Case, When, IntegerField

    when_clauses = [When(pk=pk, then=age) for pk, age in age_mapping.items()]

    return qs.annotate(age=Case(*when_clauses, output_field=IntegerField()))


def format_nevra(name=None, epoch=0, version=None, release=None, arch=None):
    """Generate Name-Epoch-Version-Release-Arch string."""
    return "{name}-{epoch}:{version}-{release}.{arch}".format(
        name=name, epoch=epoch, version=version, release=release, arch=arch
    )


def format_nvra(name=None, version=None, release=None, arch=None):
    """Generate Name-Version-Release-Arch from a package metadata."""
    return "{name}-{version}-{release}.{arch}".format(
        name=name,
        version=version,
        release=release,
        arch=arch,
    )


def format_nevra_short(name=None, epoch=0, version=None, release=None, arch=None):
    """Returns NEVRA or NVRA based on epoch."""
    if int(epoch) > 0:
        return format_nevra(name, epoch, version, release, arch)
    return format_nvra(name, version, release, arch)


def read_crpackage_from_artifact(artifact, working_dir="."):
    """
    Helper function for creating package.

    Copy file to a temp directory and parse it.

    Returns: package model as dict

    Args:
        artifact: inited and validated artifact to save
    """
    filename = f"{artifact.pulp_id}.rpm"
    artifact_file = artifact.pulp_domain.get_storage().open(artifact.file.name)
    with tempfile.NamedTemporaryFile("wb", dir=working_dir, suffix=filename) as temp_file:
        shutil.copyfileobj(artifact_file, temp_file)
        temp_file.flush()
        cr_pkginfo = cr.package_from_rpm(
            temp_file.name, changelog_limit=settings.KEEP_CHANGELOG_LIMIT
        )

    artifact_file.close()
    return cr_pkginfo


def urlpath_sanitize(*args):
    """
    Join an arbitrary number of strings into a /-separated path.

    Replaces uses of urljoin() that don't want/need urljoin's subtle semantics.

    Returns: single string provided arguments separated by single-slashes

    Args:
        Arbitrary list of arguments to be join()ed
    """
    segments = []
    for a in args + ("",):
        stripped = a.strip("/")
        if stripped:
            segments.append(stripped)
    return "/".join(segments)


def get_sha256(file_path):
    """
    Get sha256 of file.

    Args:
        file_path(string): path of file

    Returns:
        String: SHA256 of file

    """
    try:
        with open(file_path, "rb") as file_obj:
            return sha256(file_obj.read()).hexdigest()
    except FileNotFoundError:
        return None


def is_previous_version(version, target_version):
    """
    Compare version with a target version.

    Able to compare versions with integers only. Returns False for non-integer/non-1.2.3 versions.

    Args:
        version(str): version to compare
        target_version(str): version to compare with

    Returns:
        bool: True if versions are the same or if the version is older than the target version.

    """
    # if any of the versions are empty, they can't be compared and
    # a target_version need to be picked
    if version is None or target_version is None:
        return True

    # Handle equals
    if version == target_version:
        return True

    # Handle integers
    if version.isdigit() and target_version.isdigit():
        return int(version) <= int(target_version)

    # Handle 1.2.3
    version_components = version.split(".")
    target_version_components = target_version.split(".")

    if len(version_components) == len(target_version_components):
        for comp_a, comp_b in zip(version_components, target_version_components):
            if comp_a.isdigit() and comp_b.isdigit():
                # if both strings contain numeric information, convert them to ints and compare
                a = int(comp_a)
                b = int(comp_b)
                if a == b:
                    # this 'place' is equal, move to next
                    continue
                else:
                    # not equal, return comparison
                    return a < b
            else:
                # if one of the versions contains non-numeric information we cannot compare
                return False
        else:
            # All places compared, is_prev=>T if last place is equal
            return int(comp_a) == int(comp_b)
    # len wasn't equal
    return False


def parse_time(value):
    """
    Parse datetime values from a string.

    Able to distinguish between timestamp and iso format.

    Args:
        value(str): unformatted time value

    Returns:
        int | datetime | None: formatted time value
    """
    return int(value) if value.isdigit() else parse_datetime(value)


def _get_datapkg_sample_rpm_copy(basedir: str):
    sample_rpm = files("pulp_rpm").joinpath("tests/sample-rpm-0-0.x86_64.rpm")
    copy_rpm = shutil.copy(sample_rpm, basedir)
    return Path(copy_rpm)


class RpmTool:
    """
    A wrapper utility for rpm cli tool.

    Args:
        root: Alternative root directory passed to `rpm --root`
    """

    INVALID_SIGNATURE_ERROR_MSG = "Signature is invalid or pubkey is unreachable"
    UNKNOWN_ERROR_MSG = "Some unknown error occurred"
    UNSIGNED_ERROR_MSG = "The package is not signed"

    def __init__(self, root: t.Optional[Path] = None):
        completed_process = subprocess.run(
            ["which", "rpmsign"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if completed_process.returncode != 0:
            raise RuntimeError("Rpm cli tool is not installed on your system.")

        self.opts = ["--root", str(root.absolute())] if root else []

    @staticmethod
    def get_empty_rpm(basedir: str) -> Path:
        """
        Get an empty rpm package.

        Args:
            basedir: The dir where the rpm will be placed.
        """
        return _get_datapkg_sample_rpm_copy(basedir)

    def import_pubkey_file(self, pubkey_file: str):
        """
        Import public_key file (ascii-armored) into the rpm-tool.

        Args:
            import_pubkey: The public key file in ascii-armored format.
        """
        cmd = ("rpm", *self.opts, "--import", pubkey_file)
        completed_process = subprocess.run(cmd, capture_output=True)
        if completed_process.returncode != 0:
            raise RuntimeError(
                f"Could not import public key into rpm-tool:\n{completed_process.stderr.decode()}"
            )

    def import_pubkey_string(self, pubkey_content: str):
        """
        Import public_key string (ascii-armored) into the rpm-tool.

        Parameters:
            import_pubkey: The public key string in ascii-armored format.
        """
        with tempfile.NamedTemporaryFile() as pubkey_file:
            pubkey_file.write(pubkey_content.encode())
            pubkey_file.flush()
            self.import_pubkey_file(pubkey_file.name)

    def verify_signature(self, rpm_package_file: Path, raises=True):
        """
        Verify that an Rpm Package is signed by some of the imported pubkey.

        Parameters:
            rpm_package_file: Path object to rpm package

        Returns:
            True (if has valid)

        Raises:
            InvalidSignature (for invalid/unsigned package)

        Notes:
            This is based on the command: `rpm --checksig camel-0.1-1.noarch.rpm`
            Which have the following scenarios/outputs:

            - unsigned:
                * returncode: 0
                * output: "camel-0.1-1.noarch.rpm: digests OK"
            - signed, but rpm doesnt have pubkey imported:
                * returncode: 1
                * output: "camel-0.1-1.noarch.rpm: digests SIGNATURES NOT OK"
            - signed and rpm can validate:
                * returncode: 0
                * output: "camel-0.1-1.noarch.rpm: digests signatures OK"
        """
        cmd = ("rpm", *self.opts, "--checksig", str(rpm_package_file.resolve()))
        completed_process = subprocess.run(cmd, capture_output=True)
        stdout = completed_process.stdout.decode()
        stderr = completed_process.stderr.decode()
        output = f"\nstdout: {stdout}\nstderr: {stderr}"
        if completed_process.returncode != 0:
            if "SIGNATURES NOT OK" in stdout:
                raise InvalidSignatureError(f"{RpmTool.INVALID_SIGNATURE_ERROR_MSG}: {output}")
            raise TypeError(f"{RpmTool.UNKNOWN_ERROR_MSG}: {output}")
        elif "signatures" not in output:
            raise InvalidSignatureError(f"{RpmTool.UNSIGNED_ERROR_MSG}: {output}")
        return True



from django.test import TestCase

from pulp_rpm.app.tasks.publishing import PkgBuild, _CollisionManager


class TestPublishing(TestCase):
    """Test the publishing task."""

    def test_collision_manager(self):
        """Test that the collision manager is correctly choosing which packages to keep."""
        low_epoch = PkgBuild(cid=1, epoch=0, build_time=100)
        mid_epoch = PkgBuild(cid=2, epoch=1, build_time=150)
        high_epoch = PkgBuild(cid=3, epoch=2, build_time=50)
        low_build_time = PkgBuild(cid=4, epoch=0, build_time=200)
        mid_build_time = PkgBuild(cid=5, epoch=0, build_time=250)
        high_build_time = PkgBuild(cid=6, epoch=0, build_time=300)

        # Test that collisions on nevra or path are checking epoch correctly.
        cm = _CollisionManager()
        cm.add(mid_epoch, "nevra", "path", None)
        cm.add(low_epoch, "nevra", "path2", None)
        self.assertEqual([mid_epoch.cid], cm.retained_cids())
        cm.add(high_epoch, "nevra2", "path", None)
        self.assertEqual([high_epoch.cid], cm.retained_cids())

        # Test that collisions on nevra or path are checking build_time correctly.
        cm = _CollisionManager()
        cm.add(mid_build_time, "nevra", "path", None)
        cm.add(low_build_time, "nevra", "path2", None)
        self.assertEqual([mid_build_time.cid], cm.retained_cids())
        cm.add(high_build_time, "nevra2", "path", None)
        self.assertEqual([high_build_time.cid], cm.retained_cids())

        # Test that collisions on 2nd_path are being handled correctly.
        cm = _CollisionManager()
        cm.add(mid_build_time, "nevra", "path", "second_path")
        cm.add(low_build_time, "nevra2", "path2", "second_path")
        self.assertEqual({mid_build_time.cid, low_build_time.cid}, set(cm.retained_cids()))
        self.assertIn(mid_build_time.cid, cm.cid_to_second_path)
        self.assertEqual(cm.cid_to_second_path[mid_build_time.cid], "second_path")
        self.assertNotIn(low_build_time.cid, cm.cid_to_second_path)

        cm.add(high_build_time, "nevra3", "path3", "second_path")
        self.assertEqual(
            {mid_build_time.cid, low_build_time.cid, high_build_time.cid}, set(cm.retained_cids())
        )
        self.assertIn(high_build_time.cid, cm.cid_to_second_path)
        self.assertEqual(cm.cid_to_second_path[high_build_time.cid], "second_path")
        self.assertNotIn(mid_build_time.cid, cm.cid_to_second_path)



# Generated by Django 4.2.27 on 2025-12-09 19:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0065_alter_package_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rpmpublication',
            name='layout',
            field=models.TextField(choices=[('nested_alphabetically', 'nested_alphabetically'), ('flat', 'flat'), ('nested_by_digest', 'nested_by_digest'), ('nested_by_both', 'nested_by_both')], null=True),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='layout',
            field=models.TextField(choices=[('nested_alphabetically', 'nested_alphabetically'), ('flat', 'flat'), ('nested_by_digest', 'nested_by_digest'), ('nested_by_both', 'nested_by_both')], null=True),
        ),
    ]



"""
Unit tests for RPM version comparison functionality.
"""

from pulp_rpm.app.rpm_version import (
    RpmVersion,
    _compare_rpm_versions,
    _compare_version_strings,
    from_evr,
)


class TestRpmVersionComparison:
    """Test basic RPM version comparison functionality."""

    def test_evr_tostr(self):
        """Test that EVRs are printed as expected."""
        evr = RpmVersion("", "1.2.3", "45")
        assert str(evr) == "1.2.3-45"

        evr = RpmVersion("0", "1.2.3", "45")
        assert str(evr) == "0:1.2.3-45"

    def test_evr_parse(self):
        """Test that a correctly formed EVR string is parsed correctly."""
        evr = RpmVersion.from_string("1.2.3-45")
        expected = RpmVersion("", "1.2.3", "45")
        assert evr == expected

        evr = RpmVersion.from_string("0:1.2.3-45")
        expected = RpmVersion("0", "1.2.3", "45")
        assert evr == expected

        evr = RpmVersion.from_string("1:2.3.4-5")
        expected = RpmVersion("1", "2.3.4", "5")
        assert evr == expected

    def test_evr_parse_edge_cases(self):
        """Test that various not-well-formed EVR strings still get parsed in a sensible way."""
        assert from_evr("-") == ("", "", "")
        assert from_evr(".") == ("", ".", "")
        assert from_evr(":") == ("", "", "")
        assert from_evr(":-") == ("", "", "")
        assert from_evr(".-") == ("", ".", "")
        assert from_evr("0") == ("", "0", "")
        assert from_evr("0-") == ("", "0", "")
        assert from_evr(":0") == ("", "0", "")
        assert from_evr(":0-") == ("", "0", "")
        assert from_evr("0:") == ("0", "", "")
        assert from_evr("asdf:") == ("asdf", "", "")
        assert from_evr("~:") == ("~", "", "")

    def test_rpm_evr_compare(self):
        """Test direct comparison of rpm EVR strings."""
        assert _compare_rpm_versions("0:1.2.3-45", "1.2.3-45") == 0
        assert _compare_rpm_versions("1.2.3-45", "1:1.2.3-45") < 0
        assert _compare_rpm_versions("1.2.3-46", "1.2.3-45") > 0

    def test_evr_ord(self):
        """Test comparing EVRs using comparison operators."""
        # compare the same EVR without epoch as equal
        evr1 = RpmVersion.from_string("1.2.3-45")
        evr2 = RpmVersion.from_string("1.2.3-45")
        assert evr1 == evr2

        # compare the same EVR with epoch as equal
        evr1 = RpmVersion.from_string("2:1.2.3-45")
        evr2 = RpmVersion.from_string("2:1.2.3-45")
        assert evr1 == evr2

        # compare the same EVR with zero-epoch as equal to default-epoch
        evr1 = RpmVersion.from_string("1.2.3-45")
        evr2 = RpmVersion.from_string("0:1.2.3-45")
        assert evr1 == evr2

        # compare EVR with higher epoch and same version / release
        evr1 = RpmVersion.from_string("1.2.3-45")
        evr2 = RpmVersion.from_string("1:1.2.3-45")
        assert evr1 < evr2

        # compare EVR with higher epoch taken over EVR with higher version
        evr1 = RpmVersion.from_string("4.2.3-45")
        evr2 = RpmVersion.from_string("1:1.2.3-45")
        assert evr1 < evr2

        # compare EVR with higher version
        evr1 = RpmVersion.from_string("1.2.3-45")
        evr2 = RpmVersion.from_string("1.2.4-45")
        assert evr1 < evr2

        # compare EVR with higher version
        evr1 = RpmVersion.from_string("1.23.3-45")
        evr2 = RpmVersion.from_string("1.2.3-45")
        assert evr1 > evr2

        # compare EVR with higher version
        evr1 = RpmVersion.from_string("12.2.3-45")
        evr2 = RpmVersion.from_string("1.2.3-45")
        assert evr1 > evr2

        # compare EVR with higher version
        evr1 = RpmVersion.from_string("1.2.3-45")
        evr2 = RpmVersion.from_string("1.12.3-45")
        assert evr1 < evr2

        # compare versions with tilde parsing as older
        evr1 = RpmVersion.from_string("~1.2.3-45")
        evr2 = RpmVersion.from_string("1.2.3-45")
        assert evr1 < evr2

        # compare versions with tilde parsing as older
        evr1 = RpmVersion.from_string("~12.2.3-45")
        evr2 = RpmVersion.from_string("1.2.3-45")
        assert evr1 < evr2

        # compare versions with tilde parsing as older
        evr1 = RpmVersion.from_string("~12.2.3-45")
        evr2 = RpmVersion.from_string("~1.2.3-45")
        assert evr1 > evr2

        # compare versions with tilde parsing as older
        evr1 = RpmVersion.from_string("~3:12.2.3-45")
        evr2 = RpmVersion.from_string("0:1.2.3-45")
        assert evr1 < evr2

        # compare release
        evr1 = RpmVersion.from_string("1.2.3-45")
        evr2 = RpmVersion.from_string("1.2.3-46")
        assert evr1 < evr2

        # compare release
        evr1 = RpmVersion.from_string("1.2.3-45.fc39")
        evr2 = RpmVersion.from_string("1.2.3-46.fc38")
        assert evr1 < evr2

        # compare release
        evr1 = RpmVersion.from_string("1.2.3-3")
        evr2 = RpmVersion.from_string("1.2.3-10")
        assert evr1 < evr2

        # compare release
        evr1 = RpmVersion.from_string("1.2.3-3.fc40")
        evr2 = RpmVersion.from_string("1.2.3-10.fc39")
        assert evr1 < evr2

    def test_compare_version_string(self):
        """Test many different combinations of version string comparison behavior."""
        assert _compare_version_strings("1.0", "1.0") == 0
        assert _compare_version_strings("1.0", "2.0") < 0
        assert _compare_version_strings("2.0", "1.0") > 0

        assert _compare_version_strings("2.0.1", "2.0.1") == 0
        assert _compare_version_strings("2.0", "2.0.1") < 0
        assert _compare_version_strings("2.0.1", "2.0") > 0

        assert _compare_version_strings("5.0.1", "5.0.1a") < 0
        assert _compare_version_strings("5.0.1a", "5.0.1") > 0

        assert _compare_version_strings("5.0.a1", "5.0.a1") == 0
        assert _compare_version_strings("5.0.1a", "5.0.1a") == 0
        assert _compare_version_strings("5.0.a1", "5.0.a2") < 0
        assert _compare_version_strings("5.0.a2", "5.0.a1") > 0

        assert _compare_version_strings("10abc", "10.1abc") < 0
        assert _compare_version_strings("10.1abc", "10abc") > 0

        assert _compare_version_strings("8.0", "8.0.rc1") < 0
        assert _compare_version_strings("8.0.rc1", "8.0") > 0

        assert _compare_version_strings("10b2", "10a1") > 0
        assert _compare_version_strings("10a2", "10b2") < 0

        assert _compare_version_strings("6.6p1", "7.5p1") < 0
        assert _compare_version_strings("7.5p1", "6.6p1") > 0

        assert _compare_version_strings("6.5p1", "6.5p1") == 0
        assert _compare_version_strings("6.5p1", "6.5p2") < 0
        assert _compare_version_strings("6.5p2", "6.5p1") > 0
        assert _compare_version_strings("6.5p2", "6.6p1") < 0
        assert _compare_version_strings("6.6p1", "6.5p2") > 0

        assert _compare_version_strings("6.5p10", "6.5p10") == 0
        assert _compare_version_strings("6.5p1", "6.5p10") < 0
        assert _compare_version_strings("6.5p10", "6.5p1") > 0

        assert _compare_version_strings("abc10", "abc10") == 0
        assert _compare_version_strings("abc10", "abc10.1") < 0
        assert _compare_version_strings("abc10.1", "abc10") > 0

        assert _compare_version_strings("abc.4", "abc.4") == 0
        assert _compare_version_strings("abc.4", "8") < 0
        assert _compare_version_strings("8", "abc.4") > 0
        assert _compare_version_strings("abc.4", "2") < 0
        assert _compare_version_strings("2", "abc.4") > 0

        assert _compare_version_strings("1.0aa", "1.0aa") == 0
        assert _compare_version_strings("1.0a", "1.0aa") < 0
        assert _compare_version_strings("1.0aa", "1.0a") > 0

    def test_version_comparison_numeric_handling(self):
        """Test handling of numeric-like values in version strings."""
        assert _compare_version_strings("10.0001", "10.0001") == 0
        # sequences of leading zeroes are meant to be ignored - it's not *actually* treated
        # like a numeric value
        assert _compare_version_strings("10.0001", "10.1") == 0
        assert _compare_version_strings("10.1", "10.0001") == 0
        assert _compare_version_strings("10.0001", "10.0039") < 0
        assert _compare_version_strings("10.0039", "10.0001") > 0
        # but sequences of zeroes within a numeric segment are not ignored
        assert _compare_version_strings("10.1", "10.10001") < 0
        assert _compare_version_strings("10.1111", "10.10001") < 0
        assert _compare_version_strings("10.11111", "10.10001") > 0

        assert _compare_version_strings("20240521", "20240521") == 0
        assert _compare_version_strings("20240521", "20240522") < 0
        assert _compare_version_strings("20240522", "20240521") > 0
        assert _compare_version_strings("20240521", "202405210") < 0

    def test_version_comparison_tilde_and_caret(self):
        """Test behavior of tilde and caret operators."""
        assert _compare_version_strings("1.0~rc1", "1.0~rc1") == 0
        assert _compare_version_strings("1.0~rc1", "1.0") < 0
        assert _compare_version_strings("1.0", "1.0~rc1") > 0
        assert _compare_version_strings("1.0~rc1", "1.0~rc2") < 0
        assert _compare_version_strings("1.0~rc2", "1.0~rc1") > 0
        assert _compare_version_strings("1.0~rc1~git123", "1.0~rc1~git123") == 0
        assert _compare_version_strings("1.0~rc1~git123", "1.0~rc1") < 0
        assert _compare_version_strings("1.0~rc1", "1.0~rc1~git123") > 0

        assert _compare_version_strings("1.0^", "1.0^") == 0
        assert _compare_version_strings("1.0", "1.0^") < 0
        assert _compare_version_strings("1.0^", "1.0") > 0

        assert _compare_version_strings("1.0", "1.0git1^") < 0
        assert _compare_version_strings("1.0^git1", "1.0^git2") < 0
        assert _compare_version_strings("1.01", "1.0^git1") > 0
        assert _compare_version_strings("1.0^20240501", "1.0^20240501") == 0
        assert _compare_version_strings("1.0^20240501", "1.0.1") < 0
        assert _compare_version_strings("1.0^20240501^git1", "1.0^20240501^git1") == 0
        assert _compare_version_strings("1.0^20240502", "1.0^20240501^git1") > 0
        assert _compare_version_strings("1.0~rc1^git1", "1.0~rc1^git1") == 0
        assert _compare_version_strings("1.0~rc1", "1.0~rc1^git1") < 0
        assert _compare_version_strings("1.0~rc1^git1", "1.0~rc1") > 0
        assert _compare_version_strings("1.0^git1~pre", "1.0^git1~pre") == 0
        assert _compare_version_strings("1.0^git1~pre", "1.0^git1") < 0
        assert _compare_version_strings("1.0^git1", "1.0^git1~pre") > 0

    def test_non_intuitive_comparison_behavior(self):
        """Test some version comparison behavior that is a bit non-intuitive."""
        # (but needs to be maintained for compatibility)
        assert _compare_version_strings("1e.fc33", "1.fc33") < 0
        assert _compare_version_strings("1g.fc33", "1.fc33") > 0

    def test_non_alphanumeric_equivalence(self):
        """Test handling of non-alphanumeric ascii characters (excluding separators)."""
        # the existence of sequences of non-alphanumeric characters should not impact
        # the version comparison at all
        assert _compare_version_strings("b", "b") == 0
        assert _compare_version_strings("b+", "b+") == 0
        assert _compare_version_strings("b+", "b_") == 0
        assert _compare_version_strings("b_", "b+") == 0
        assert _compare_version_strings("+b", "+b") == 0
        assert _compare_version_strings("+b", "_b") == 0
        assert _compare_version_strings("_b", "+b") == 0

        assert _compare_version_strings("+b", "++b") == 0
        assert _compare_version_strings("+b", "+b+") == 0

        assert _compare_version_strings("+.", "+_") == 0
        assert _compare_version_strings("_+", "+.") == 0
        assert _compare_version_strings("+", ".") == 0
        assert _compare_version_strings(",", "+") == 0

        assert _compare_version_strings("++", "_") == 0
        assert _compare_version_strings("+", "..") == 0

        assert _compare_version_strings("4_0", "4_0") == 0
        assert _compare_version_strings("4_0", "4.0") == 0
        assert _compare_version_strings("4.0", "4_0") == 0

        assert _compare_version_strings("4.999", "5.0") < 0
        assert _compare_version_strings("4.999.9", "5.0") < 0
        assert _compare_version_strings("5.0", "4.999_9") > 0

        # except when it comes to breaking up sequences of alphanumeric characters
        # that do impact the comparison
        assert _compare_version_strings("4.999", "4.999.9") < 0
        assert _compare_version_strings("4.999", "4.99.9") > 0

    def test_non_ascii_character_equivalence(self):
        """Test handling of non-ascii characters."""
        # the existence of sequences of non-ascii characters should not impact the
        # version comparison at all
        assert _compare_version_strings("1.1.Á.1", "1.1.1") == 0
        assert _compare_version_strings("1.1.Á", "1.1.Á") == 0
        assert _compare_version_strings("1.1.Á", "1.1.Ê") == 0
        assert _compare_version_strings("1.1.ÁÁ", "1.1.Á") == 0
        assert _compare_version_strings("1.1.Á", "1.1.ÊÊ") == 0

        # except when it comes to breaking up sequences of ascii characters that do
        # impact the comparison
        assert _compare_version_strings("1.1Á1", "1.11") < 0



# Generated by Django 4.2.23 on 2025-08-13 12:47

from django.db import migrations, models
import django.db.models.deletion
import pulp_rpm.app.models.package
import pulpcore.app.models.base
import pulpcore.app.util


class Migration(migrations.Migration):

    replaces = [('rpm', '0001_initial'), ('rpm', '0002_updaterecord_reboot_suggested'), ('rpm', '0003_DATA_incorrect_json'), ('rpm', '0004_add_metadata_signing_service_fk'), ('rpm', '0005_optimize_sync'), ('rpm', '0006_opensuse_support'), ('rpm', '0007_checksum_types'), ('rpm', '0008_advisory_pkg_sumtype_as_int'), ('rpm', '0009_revision_null'), ('rpm', '0010_revision_null_redo'), ('rpm', '0011_rpmremote_sles_auth_token'), ('rpm', '0012_remove_pkg_group_env_cat_related_pkgs'), ('rpm', '0013_RAW_rpm_evr_extension'), ('rpm', '0014_rpmrepository_package_retention_policy'), ('rpm', '0015_repo_metadata'), ('rpm', '0016_dist_tree_nofk'), ('rpm', '0017_merge_advisory_collections'), ('rpm', '0018_updatecollection__update_record'), ('rpm', '0019_migrate_updatecollection_data'), ('rpm', '0020_remove_updatecollection_m2m'), ('rpm', '0021_rename_updatecollection_update_record'), ('rpm', '0022_add_collections_related_name'), ('rpm', '0023_increase_distribution_release_short'), ('rpm', '0024_change_subrepo_relation_properties'), ('rpm', '0025_remove_orphaned_subrepos'), ('rpm', '0026_add_gpgcheck_options'), ('rpm', '0027_checksum_null'), ('rpm', '0028_rpmrepository_last_sync_repomd_cheksum'), ('rpm', '0029_rpmpublication_sqlite_metadata'), ('rpm', '0030_DATA_fix_updaterecord'), ('rpm', '0031_modulemd_static_context'), ('rpm', '0032_ulnremote'), ('rpm', '0033_new_distribution_model'), ('rpm', '0034_auto_publish'), ('rpm', '0035_fix_auto_publish'), ('rpm', '0036_checksum_type'), ('rpm', '0037_DATA_remove_rpmrepository_sub_repo'), ('rpm', '0037_update_json_field'), ('rpm', '0038_fix_sync_optimization'), ('rpm', '0039_disttree_digest'), ('rpm', '0040_rpmalternatecontentsource'), ('rpm', '0041_modulemdobsolete'), ('rpm', '0042_alter_repometadatafile_data_type'), ('rpm', '0043_textfield_conversion'), ('rpm', '0044_noartifact_modules'), ('rpm', '0045_modulemd_fields'), ('rpm', '0046_rbac_perms'), ('rpm', '0047_modulemd_datefield'), ('rpm', '0048_artifacts_dependencies_fix'), ('rpm', '0049_profiles_fix'), ('rpm', '0050_alter_addon_pulp_id_alter_checksum_pulp_id_and_more'), ('rpm', '0051_alter_distributiontree_unique_together_and_more'), ('rpm', '0052_modulemd_digest'), ('rpm', '0053_rpmdistribution_generate_repo_config'), ('rpm', '0054_remove_gpg_fields'), ('rpm', '0055_add_repo_config_field'), ('rpm', '0056_remove_rpmpublication_sqlite_metadata_and_more'), ('rpm', '0057_rpmpublication_checksum_type_and_more'), ('rpm', '0058_alter_addon_repository_alter_variant_repository'), ('rpm', '0059_rpmpublication_compression_type_and_more'), ('rpm', '0060_rpmpublication_compression_type_empty'), ('rpm', '0061_fix_modulemd_defaults_digest'), ('rpm', '0062_rpmpackagesigningservice_and_more')]

    initial = True

    dependencies = [
        ('core', '0091_systemid'),
        ('core', '0106_alter_artifactdistribution_distribution_ptr_and_more'),
        ('core', '0114_remove_task_args_remove_task_kwargs'),
    ]

    operations = [
        migrations.CreateModel(
            name='DistributionTree',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('header_version', models.TextField()),
                ('release_name', models.TextField()),
                ('release_short', models.TextField()),
                ('release_version', models.TextField()),
                ('release_is_layered', models.BooleanField(default=False)),
                ('base_product_name', models.TextField(null=True)),
                ('base_product_short', models.TextField(null=True)),
                ('base_product_version', models.TextField(null=True)),
                ('arch', models.TextField()),
                ('build_timestamp', models.FloatField()),
                ('instimage', models.TextField(null=True)),
                ('mainimage', models.TextField(null=True)),
                ('discnum', models.IntegerField(null=True)),
                ('totaldiscs', models.IntegerField(null=True)),
                ('digest', models.TextField()),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='Package',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('name', models.CharField(max_length=255)),
                ('epoch', models.CharField(max_length=10)),
                ('version', models.CharField(max_length=255)),
                ('release', models.CharField(max_length=255)),
                ('arch', models.CharField(max_length=20)),
                ('pkgId', models.TextField(db_index=True)),
                ('checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')])),
                ('summary', models.TextField()),
                ('description', models.TextField()),
                ('url', models.TextField()),
                ('changelogs', models.JSONField(default=list)),
                ('files', models.JSONField(default=list)),
                ('requires', models.JSONField(default=list)),
                ('provides', models.JSONField(default=list)),
                ('conflicts', models.JSONField(default=list)),
                ('obsoletes', models.JSONField(default=list)),
                ('suggests', models.JSONField(default=list)),
                ('enhances', models.JSONField(default=list)),
                ('recommends', models.JSONField(default=list)),
                ('supplements', models.JSONField(default=list)),
                ('location_base', models.TextField()),
                ('location_href', models.TextField()),
                ('rpm_buildhost', models.TextField()),
                ('rpm_group', models.TextField()),
                ('rpm_license', models.TextField()),
                ('rpm_packager', models.TextField()),
                ('rpm_sourcerpm', models.TextField()),
                ('rpm_vendor', models.TextField()),
                ('rpm_header_start', models.BigIntegerField(null=True)),
                ('rpm_header_end', models.BigIntegerField(null=True)),
                ('is_modular', models.BooleanField(default=False)),
                ('size_archive', models.BigIntegerField(null=True)),
                ('size_installed', models.BigIntegerField(null=True)),
                ('size_package', models.BigIntegerField(null=True)),
                ('time_build', models.BigIntegerField(null=True)),
                ('time_file', models.BigIntegerField(null=True)),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'name', 'epoch', 'version', 'release', 'arch', 'checksum_type', 'pkgId')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='UpdateRecord',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('id', models.TextField(db_index=True)),
                ('updated_date', models.TextField(null=True)),
                ('description', models.TextField()),
                ('issued_date', models.TextField()),
                ('fromstr', models.TextField()),
                ('status', models.TextField()),
                ('title', models.TextField()),
                ('summary', models.TextField()),
                ('version', models.TextField()),
                ('type', models.TextField()),
                ('severity', models.TextField()),
                ('solution', models.TextField()),
                ('release', models.TextField()),
                ('rights', models.TextField()),
                ('pushcount', models.TextField(blank=True)),
                ('digest', models.TextField(db_index=True)),
                ('reboot_suggested', models.BooleanField(default=False)),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageGroup',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('id', models.TextField()),
                ('default', models.BooleanField(default=False)),
                ('user_visible', models.BooleanField(default=False)),
                ('display_order', models.IntegerField(null=True)),
                ('name', models.TextField()),
                ('description', models.TextField(default='')),
                ('packages', models.JSONField(default=list)),
                ('biarch_only', models.BooleanField(default=False)),
                ('desc_by_lang', models.JSONField(default=dict)),
                ('name_by_lang', models.JSONField(default=dict)),
                ('digest', models.TextField(db_index=True)),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='UpdateCollection',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.TextField(null=True)),
                ('shortname', models.TextField(null=True)),
                ('module', models.JSONField(null=True)),
                ('update_record', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='collections', to='rpm.updaterecord')),
            ],
            options={
                'abstract': False,
                'unique_together': {('name', 'update_record')},
            },
        ),
        migrations.CreateModel(
            name='Checksum',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('path', models.TextField()),
                ('checksum', models.TextField()),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='checksums', to='rpm.distributiontree')),
            ],
            options={
                'unique_together': {('path', 'checksum', 'distribution_tree')},
            },
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.TextField()),
                ('path', models.TextField()),
                ('platforms', models.TextField()),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images', to='rpm.distributiontree')),
            ],
            options={
                'unique_together': {('name', 'path', 'platforms', 'distribution_tree')},
            },
        ),
        migrations.CreateModel(
            name='RpmAlternateContentSource',
            fields=[
                ('alternatecontentsource_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.alternatecontentsource')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'permissions': [('refresh_rpmalternatecontentsource', 'Refresh an Alternate Content Source'), ('manage_roles_rpmalternatecontentsource', 'Can manage roles on ACS')],
            },
            bases=('core.alternatecontentsource',),
        ),
        migrations.CreateModel(
            name='RpmRemote',
            fields=[
                ('remote_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.remote')),
                ('sles_auth_token', models.TextField(null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'permissions': [('manage_roles_rpmremote', 'Can manage roles on an RPM remotes')],
            },
            bases=('core.remote',),
        ),
        migrations.CreateModel(
            name='UlnRemote',
            fields=[
                ('remote_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.remote')),
                ('uln_server_base_url', models.TextField(null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'permissions': [('manage_roles_ulnremote', 'Can manage roles on an ULN remotes')],
            },
            bases=('core.remote',),
        ),
        migrations.CreateModel(
            name='UpdateCollectionPackage',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('arch', models.TextField()),
                ('epoch', models.TextField()),
                ('filename', models.TextField()),
                ('name', models.TextField()),
                ('reboot_suggested', models.BooleanField(default=False)),
                ('release', models.TextField()),
                ('src', models.TextField()),
                ('sum', models.TextField()),
                ('version', models.TextField()),
                ('update_collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='packages', to='rpm.updatecollection')),
                ('relogin_suggested', models.BooleanField(default=False)),
                ('restart_suggested', models.BooleanField(default=False)),
                ('sum_type', models.PositiveIntegerField(choices=[(0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 6), (7, 7)], default=None, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UpdateReference',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('href', models.TextField()),
                ('ref_id', models.TextField(null=True)),
                ('title', models.TextField(null=True)),
                ('ref_type', models.TextField()),
                ('update_record', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='references', to='rpm.updaterecord')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ModulemdDefaults',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('module', models.TextField()),
                ('stream', models.TextField()),
                ('profiles', models.JSONField(default=list)),
                ('digest', models.TextField(db_index=True)),
                ('snippet', models.TextField()),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='ModulemdObsolete',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('modified', models.TextField()),
                ('module_name', models.TextField()),
                ('module_stream', models.TextField()),
                ('message', models.TextField()),
                ('override_previous', models.BooleanField(null=True)),
                ('module_context', models.TextField(null=True)),
                ('eol_date', models.TextField(null=True)),
                ('obsoleted_by_module_name', models.TextField(null=True)),
                ('obsoleted_by_module_stream', models.TextField(null=True)),
                ('snippet', models.TextField()),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'modified', 'module_name', 'module_stream')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageCategory',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('id', models.TextField()),
                ('name', models.TextField()),
                ('description', models.TextField(default='')),
                ('display_order', models.IntegerField(null=True)),
                ('group_ids', models.JSONField(default=list)),
                ('desc_by_lang', models.JSONField(default=dict)),
                ('name_by_lang', models.JSONField(default=dict)),
                ('digest', models.TextField(db_index=True)),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageEnvironment',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('id', models.TextField()),
                ('name', models.TextField()),
                ('description', models.TextField(default='')),
                ('display_order', models.IntegerField(null=True)),
                ('group_ids', models.JSONField(default=list)),
                ('option_ids', models.JSONField(default=list)),
                ('desc_by_lang', models.JSONField(default=dict)),
                ('name_by_lang', models.JSONField(default=dict)),
                ('digest', models.TextField(db_index=True)),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageLangpacks',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('matches', models.JSONField(default=dict)),
                ('digest', models.TextField(db_index=True)),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='RepoMetadataFile',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('data_type', models.TextField()),
                ('checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')])),
                ('checksum', models.TextField()),
                ('relative_path', models.TextField()),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'data_type', 'checksum', 'relative_path')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='Modulemd',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content')),
                ('name', models.TextField()),
                ('stream', models.TextField()),
                ('version', models.TextField()),
                ('context', models.TextField()),
                ('arch', models.TextField()),
                ('dependencies', models.JSONField(default=list)),
                ('artifacts', models.JSONField(default=list)),
                ('packages', models.ManyToManyField(to='rpm.package')),
                ('static_context', models.BooleanField(null=True)),
                ('snippet', models.TextField()),
                ('description', models.TextField()),
                ('profiles', models.JSONField(default=dict)),
                ('_pulp_domain', models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain')),
                ('digest', models.TextField()),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('_pulp_domain', 'name', 'stream', 'version', 'context', 'arch', 'digest')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='RpmDistribution',
            fields=[
                ('distribution_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.distribution')),
                ('generate_repo_config', models.BooleanField(default=False)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'permissions': [('manage_roles_rpmdistribution', 'Can manage roles on an RPM distribution')],
            },
            bases=('core.distribution',),
        ),
        migrations.CreateModel(
            name='Addon',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('addon_id', models.TextField()),
                ('uid', models.TextField()),
                ('name', models.TextField()),
                ('type', models.TextField()),
                ('packages', models.TextField()),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='addons', to='rpm.distributiontree')),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, related_name='addons', to='core.repository')),
            ],
            options={
                'unique_together': {('addon_id', 'uid', 'name', 'type', 'packages', 'distribution_tree')},
            },
        ),
        migrations.CreateModel(
            name='Variant',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('variant_id', models.TextField()),
                ('uid', models.TextField()),
                ('name', models.TextField()),
                ('type', models.TextField()),
                ('packages', models.TextField()),
                ('source_packages', models.TextField(null=True)),
                ('source_repository', models.TextField(null=True)),
                ('debug_packages', models.TextField(null=True)),
                ('debug_repository', models.TextField(null=True)),
                ('identity', models.TextField(null=True)),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variants', to='rpm.distributiontree')),
                ('repository', models.ForeignKey(null=True, on_delete=django.db.models.deletion.RESTRICT, related_name='variants', to='core.repository')),
            ],
            options={
                'unique_together': {('variant_id', 'uid', 'name', 'type', 'packages', 'distribution_tree')},
            },
        ),
        migrations.CreateModel(
            name='RpmPublication',
            fields=[
                ('publication_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.publication')),
                ('metadata_checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')])),
                ('package_checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')])),
                ('repo_config', models.JSONField(default=dict)),
                ('checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')])),
                ('compression_type', models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz')], null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'permissions': [('manage_roles_rpmpublication', 'Can manage roles on an RPM publication')],
            },
            bases=('core.publication',),
        ),
        migrations.CreateModel(
            name='RpmPackageSigningService',
            fields=[
                ('signingservice_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.signingservice')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.signingservice',),
        ),
        migrations.CreateModel(
            name='RpmRepository',
            fields=[
                ('repository_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.repository')),
                ('metadata_signing_service', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.asciiarmoreddetachedsigningservice')),
                ('original_checksum_types', models.JSONField(default=dict)),
                ('retain_package_versions', models.PositiveIntegerField(default=0)),
                ('autopublish', models.BooleanField(default=False)),
                ('metadata_checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True)),
                ('package_checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True)),
                ('last_sync_details', models.JSONField(default=dict)),
                ('repo_config', models.JSONField(default=dict)),
                ('checksum_type', models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True)),
                ('compression_type', models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz')], null=True)),
                ('package_signing_fingerprint', models.TextField(max_length=40, null=True)),
                ('package_signing_service', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='rpm.rpmpackagesigningservice')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'permissions': [('manage_roles_rpmrepository', 'Can manage roles on RPM repositories'), ('modify_content_rpmrepository', 'Add content to, or remove content from a repository'), ('repair_rpmrepository', 'Copy a repository'), ('sync_rpmrepository', 'Sync a repository'), ('delete_rpmrepository_version', 'Delete a repository version')],
            },
            bases=('core.repository',),
        ),
        migrations.RunSQL(
            sql="\ncreate type pulp_evr_array_item as (\n  n       NUMERIC,\n  s       TEXT\n);\n\ncreate type pulp_evr_t as (\n  epoch INT,\n  version pulp_evr_array_item[],\n  release pulp_evr_array_item[]\n);\n\nCREATE OR REPLACE FUNCTION pulp_evr_trigger() RETURNS trigger AS $$\n  BEGIN\n    NEW.evr = (select ROW(coalesce(NEW.epoch::numeric,0),\n                          pulp_rpmver_array(coalesce(NEW.version,'pulp_isempty'))::pulp_evr_array_item[],\n                          pulp_rpmver_array(coalesce(NEW.release,'pulp_isempty'))::pulp_evr_array_item[])::pulp_evr_t);\n    RETURN NEW;\n  END;\n$$ language 'plpgsql';\n\ncreate or replace FUNCTION pulp_isempty(t TEXT)\n  RETURNS BOOLEAN as $$\n  BEGIN\n    return t ~ '^[[:space:]]*$';\n  END;\n$$ language 'plpgsql';\n\ncreate or replace FUNCTION pulp_isalphanum(ch CHAR)\n  RETURNS BOOLEAN as $$\n  BEGIN\n    if ascii(ch) between ascii('a') and ascii('z') or\n      ascii(ch) between ascii('A') and ascii('Z') or\n      ascii(ch) between ascii('0') and ascii('9')\n    then\n      return TRUE;\n    end if;\n    return FALSE;\n  END;\n$$ language 'plpgsql';\n\ncreate or replace function pulp_isdigit(ch CHAR)\n  RETURNS BOOLEAN as $$\n  BEGIN\n    if ascii(ch) between ascii('0') and ascii('9')\n    then\n    return TRUE;\n    end if;\n    return FALSE;\n  END ;\n$$ language 'plpgsql';\n\ncreate or replace FUNCTION pulp_rpmver_array (string1 IN VARCHAR)\n  RETURNS pulp_evr_array_item[] as $$\n  declare\n    str1 VARCHAR := string1;\n    digits VARCHAR(10) := '0123456789';\n    lc_alpha VARCHAR(27) := 'abcdefghijklmnopqrstuvwxyz';\n    uc_alpha VARCHAR(27) := 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';\n    alpha VARCHAR(54) := lc_alpha || uc_alpha;\n    one VARCHAR;\n    isnum BOOLEAN;\n    ver_array pulp_evr_array_item[] := ARRAY[]::pulp_evr_array_item[];\n  BEGIN\n    if str1 is NULL\n    then\n      RAISE EXCEPTION 'VALUE_ERROR.';\n    end if;\n\n    one := str1;\n    <<segment_loop>>\n    while one <> ''\n    loop\n      declare\n        segm1 VARCHAR;\n        segm1_n NUMERIC := 0;\n      begin\n        -- Throw out all non-alphanum characters\n        while one <> '' and not pulp_isalphanum(one)\n        loop\n          one := substr(one, 2);\n        end loop;\n        str1 := one;\n        if str1 <> '' and pulp_isdigit(str1)\n        then\n          str1 := ltrim(str1, digits);\n          isnum := true;\n        else\n          str1 := ltrim(str1, alpha);\n          isnum := false;\n        end if;\n        if str1 <> ''\n        then segm1 := substr(one, 1, length(one) - length(str1));\n        else segm1 := one;\n        end if;\n\n        if segm1 = '' then return ver_array; end if; /* arbitrary */\n        if isnum\n        then\n          segm1 := ltrim(segm1, '0');\n          if segm1 <> '' then segm1_n := segm1::numeric; end if;\n          segm1 := NULL;\n        else\n        end if;\n        ver_array := array_append(ver_array, (segm1_n, segm1)::pulp_evr_array_item);\n        one := str1;\n      end;\n    end loop segment_loop;\n\n    return ver_array;\n  END ;\n$$ language 'plpgsql';\n",
        ),
        migrations.RunSQL(
            sql='\n-- create pulp_evr_t on insert, so it matches the provided E/V/R cols\nCREATE TRIGGER pulp_evr_insert_trigger\n  BEFORE INSERT\n  ON rpm_package\n  FOR EACH ROW\n  EXECUTE PROCEDURE pulp_evr_trigger();\n\n-- create pulp_evr_t on update, so it continues to match the provided E/V/R cols, but only if Something Changed\nCREATE TRIGGER pulp_evr_update_trigger\n  BEFORE UPDATE OF epoch, version, release\n  ON rpm_package\n  FOR EACH ROW\n  WHEN (\n    OLD.epoch IS DISTINCT FROM NEW.epoch OR\n    OLD.version IS DISTINCT FROM NEW.version OR\n    OLD.release IS DISTINCT FROM NEW.release\n  )\n  EXECUTE PROCEDURE pulp_evr_trigger();\n',
        ),
        migrations.AddField(
            model_name='package',
            name='evr',
            field=pulp_rpm.app.models.package.RpmVersionField(),
        ),
        migrations.RunSQL(
            sql='\nupdate rpm_package set evr = (\n  select ROW(coalesce(epoch::numeric,0),\n             pulp_rpmver_array(version)::pulp_evr_array_item[],\n             pulp_rpmver_array(release)::pulp_evr_array_item[])::pulp_evr_t\n  );\n',
        ),
    ]
